from typing import Sequence
import copy

from yaml.nodes import Node, MappingNode, SequenceNode


from .koreo_semantics import ALL
from .semantics import (
    FieldIndexFn,
    Modifier,
    NodeDiagnostic,
    Position,
    Range,
    SemanticBlock,
    SemanticNode,
    SemanticStructure,
    Severity,
    compute_abs_position,
)

from . import cel_semantics


def extract_semantic_structure_info(
    anchor_abs_start: Position,
    last_token_abs_start: Position,
    yaml_node: Node,
    doc,
    semantic_type: dict[str, SemanticStructure] | SemanticStructure | None,
) -> tuple[Sequence[SemanticBlock | SemanticNode], Position]:
    match semantic_type:
        case SemanticStructure():
            clean_semantic_type = semantic_type
        case None:
            clean_semantic_type = SemanticStructure()
        case dict():
            clean_semantic_type = SemanticStructure(sub_structure=semantic_type)

    if isinstance(yaml_node, MappingNode):
        return _extract_map_structure_info(
            anchor_abs_start=anchor_abs_start,
            last_token_abs_start=last_token_abs_start,
            yaml_node=yaml_node,
            doc=doc,
            semantic_type=clean_semantic_type.sub_structure,
            strict_sub_structure_keys=clean_semantic_type.strict_sub_structure_keys,
            field_index_key_fn=clean_semantic_type.field_index_key_fn,
        )

    if isinstance(yaml_node, SequenceNode):
        return _extract_list_structure_info(
            anchor_abs_start=anchor_abs_start,
            last_token_abs_start=last_token_abs_start,
            yaml_node=yaml_node,
            doc=doc,
            semantic_type=clean_semantic_type.sub_structure,
        )

    return _extract_value_semantic_info(
        anchor_abs_start=anchor_abs_start,
        last_token_abs_start=last_token_abs_start,
        yaml_node=yaml_node,
        doc=doc,
        semantic_type=clean_semantic_type,
    )


def _extract_map_structure_info(
    anchor_abs_start: Position,
    last_token_abs_start: Position,
    yaml_node: Node,
    doc,
    semantic_type: dict[str, SemanticStructure] | SemanticStructure | None,
    strict_sub_structure_keys: bool = False,
    field_index_key_fn: FieldIndexFn | None = None,
) -> tuple[list[SemanticNode], Position]:
    semantic_nodes = []
    new_last_start = last_token_abs_start
    seen_keys = set[str]()

    semantic_type_map: dict[str, SemanticStructure] = {}
    if isinstance(semantic_type, dict):
        semantic_type_map = semantic_type
    elif semantic_type:
        semantic_type_map = {ALL: semantic_type}

    if field_index_key_fn and (index := field_index_key_fn(value=yaml_node.value)):
        sub_index_field, sub_index_value = index
    else:
        sub_index_field = None
        sub_index_value = None

    for key, value in yaml_node.value:
        node_diagnostic = None
        if f"{key.value}" in seen_keys:
            node_diagnostic = NodeDiagnostic(
                message="Duplicate key", severity=Severity.error
            )
        else:
            seen_keys.add(f"{key.value}")

        if (
            strict_sub_structure_keys
            and not node_diagnostic
            and f"{key.value}" not in semantic_type_map
        ):
            node_diagnostic = NodeDiagnostic(
                message="Unknown key", severity=Severity.error
            )

        key_semantic_type = semantic_type_map.get(key.value)
        if not key_semantic_type:
            key_semantic_type = semantic_type_map.get(ALL, SemanticStructure())

        if not key_semantic_type.type:
            key_semantic_type.type = "keyword"

        key_semantic_nodes, new_last_start = _extract_value_semantic_info(
            anchor_abs_start=anchor_abs_start,
            last_token_abs_start=new_last_start,
            yaml_node=key,
            doc=doc,
            semantic_type=key_semantic_type,
            diagnostic=node_diagnostic,
        )

        # This should never happen, in any case I am aware of.
        if len(key_semantic_nodes) != 1:
            raise Exception(f"More than one key node! {key_semantic_nodes}")

        match key_semantic_type.sub_structure:
            case SemanticStructure():
                value_semantic_type = key_semantic_type.sub_structure
            case None:
                value_semantic_type = SemanticStructure(
                    strict_sub_structure_keys=key_semantic_type.strict_sub_structure_keys
                )
            case dict():
                value_semantic_type = SemanticStructure(
                    sub_structure=key_semantic_type.sub_structure,
                    strict_sub_structure_keys=key_semantic_type.strict_sub_structure_keys,
                )

        if (
            sub_index_field
            and sub_index_field == key.value
            and not value_semantic_type.index_key_fn
        ):
            value_semantic_type = copy.replace(
                value_semantic_type, index_key_fn=lambda value: sub_index_value
            )

        value_semantic_nodes, new_last_start = _extract_value_semantic_info(
            anchor_abs_start=anchor_abs_start,
            last_token_abs_start=new_last_start,
            yaml_node=value,
            doc=doc,
            semantic_type=value_semantic_type,
        )

        key_semantic_node = key_semantic_nodes[-1]
        semantic_nodes.append(key_semantic_node._replace(children=value_semantic_nodes))

    return semantic_nodes, new_last_start


def _extract_list_structure_info(
    anchor_abs_start: Position,
    last_token_abs_start: Position,
    yaml_node: Node,
    doc,
    semantic_type: dict[str, SemanticStructure] | SemanticStructure | None,
) -> tuple[list[SemanticNode], Position]:
    semantic_nodes = []
    new_last_start = last_token_abs_start

    if isinstance(semantic_type, SemanticStructure):
        item_semantic_type = semantic_type
    else:
        item_semantic_type = SemanticStructure(sub_structure=semantic_type)

    for value in yaml_node.value:
        value_semantic_nodes, new_last_start = _extract_value_semantic_info(
            anchor_abs_start=anchor_abs_start,
            last_token_abs_start=new_last_start,
            yaml_node=value,
            doc=doc,
            semantic_type=item_semantic_type,
        )
        semantic_nodes.extend(value_semantic_nodes)

    return semantic_nodes, new_last_start


def _extract_cel_semantic_info(
    anchor_abs_start: Position,
    last_token_abs_start: Position,
    yaml_node: Node,
    doc,
    modifier: list[Modifier] | None = None,
    diagnostic: NodeDiagnostic | None = None,
):
    last_line = last_token_abs_start.line
    last_column = last_token_abs_start.character

    node_line = yaml_node.start_mark.line
    node_column = yaml_node.start_mark.column

    nodes = []

    if node_line == yaml_node.end_mark.line:
        # Single line expression.

        line_data = doc.lines[node_line]
        line_len = len(line_data)

        # This is to address lines that are quoted.
        # The quotes throw off the column position, but are not represented
        # in the value.
        eq_char_offset = yaml_node.start_mark.column
        while True:
            if line_data[eq_char_offset] == "=":
                break

            if eq_char_offset >= line_len:
                break

            eq_char_offset += 1

        seed_line = node_line - last_line
        char_offset = eq_char_offset - (0 if node_line > last_line else last_column)

        nodes.extend(
            cel_semantics.parse(
                cel_expression=[yaml_node.value],
                anchor_base_pos=_compute_rel_position(
                    line=node_line - seed_line,
                    character=node_column,
                    relative_to=anchor_abs_start,
                ),
                seed_line=seed_line,
                seed_offset=char_offset,
                abs_offset=eq_char_offset - char_offset,
            )
        )
    else:
        # Multiline expression

        # The multiline indicator character
        line_data = doc.lines[node_line]
        line_len = len(line_data)
        nodes.append(
            SemanticNode(
                position=_compute_rel_position(
                    line=node_line,
                    character=node_column,
                    relative_to=last_token_abs_start,
                ),
                anchor_rel=_compute_rel_position(
                    line=node_line, character=node_column, relative_to=anchor_abs_start
                ),
                length=line_len - node_column,
                node_type="operator",
                modifier=modifier,
                diagnostic=diagnostic,
            )
        )

        nodes.extend(
            cel_semantics.parse(
                cel_expression=doc.lines[node_line + 1 : yaml_node.end_mark.line],
                anchor_base_pos=_compute_rel_position(
                    line=node_line, character=node_column, relative_to=anchor_abs_start
                ),
                seed_line=1,
                seed_offset=0,
            )
        )

    # Go to the deepest node
    last_node = nodes[-1]
    while last_node and last_node.children:
        last_node = last_node.children[-1]

    last_abs_position = compute_abs_position(
        rel_position=last_node.anchor_rel, abs_position=anchor_abs_start
    )

    block = [
        SemanticBlock(
            local_key=None,
            index_key=None,
            anchor_rel=Range(
                start=_compute_rel_position(
                    line=yaml_node.start_mark.line,
                    character=yaml_node.start_mark.column,
                    relative_to=anchor_abs_start,
                ),
                end=_compute_rel_position(
                    line=yaml_node.end_mark.line,
                    character=yaml_node.end_mark.column,
                    relative_to=anchor_abs_start,
                ),
            ),
            children=nodes,
        )
    ]

    return (block, last_abs_position)


def _extract_scalar_semantic_info(
    anchor_abs_start: Position,
    last_token_abs_start: Position,
    yaml_node: Node,
    doc,
    semantic_type: SemanticStructure,
    diagnostic: NodeDiagnostic | None = None,
):
    last_line = last_token_abs_start.line
    last_column = last_token_abs_start.character

    node_line = yaml_node.start_mark.line
    node_column = yaml_node.start_mark.column

    if node_line == yaml_node.end_mark.line:
        value_len = yaml_node.end_mark.column - node_column
    else:
        line_data = doc.lines[node_line]
        value_len = len(line_data) - node_column

    char_offset = node_column - (0 if node_line > last_line else last_column)

    if semantic_type.local_key_fn:
        local_key = semantic_type.local_key_fn(value=yaml_node.value)
    else:
        local_key = None

    if semantic_type.index_key_fn:
        index_key = semantic_type.index_key_fn(value=yaml_node.value)
    else:
        index_key = None

    nodes = []
    while True:
        nodes.append(
            SemanticNode(
                local_key=local_key,
                index_key=index_key,
                position=Position(
                    line=node_line - last_line,
                    character=char_offset,
                ),
                anchor_rel=_compute_rel_position(
                    line=node_line,
                    character=node_column,
                    relative_to=anchor_abs_start,
                ),
                length=value_len,
                node_type=semantic_type.type,
                modifier=semantic_type.modifier,
                diagnostic=diagnostic,
            )
        )

        if node_line + 1 >= yaml_node.end_mark.line:
            break

        last_line = node_line
        last_column = node_column

        node_line += 1
        node_column = 0

        line_data = doc.lines[node_line]
        char_offset = len(line_data) - len(line_data.lstrip())
        value_len = len(line_data.strip())

    last_token_pos = Position(line=node_line, character=node_column)

    if len(nodes) <= 1:
        return (
            nodes,
            last_token_pos,
        )

    block = [
        SemanticBlock(
            local_key=local_key,
            index_key=index_key,
            anchor_rel=Range(
                start=_compute_rel_position(
                    line=yaml_node.start_mark.line,
                    character=yaml_node.start_mark.column,
                    relative_to=anchor_abs_start,
                ),
                end=_compute_rel_position(
                    line=yaml_node.end_mark.line,
                    character=yaml_node.end_mark.column,
                    relative_to=anchor_abs_start,
                ),
            ),
            children=nodes,
        )
    ]
    return (block, last_token_pos)


def _extract_value_semantic_info(
    anchor_abs_start: Position,
    last_token_abs_start: Position,
    yaml_node: Node,
    doc,
    semantic_type: dict[str, SemanticStructure] | SemanticStructure | None,
    diagnostic: NodeDiagnostic | None = None,
) -> tuple[Sequence[SemanticBlock | SemanticNode], Position]:
    match semantic_type:
        case SemanticStructure():
            clean_semantic_type = semantic_type
        case _:
            clean_semantic_type = SemanticStructure()

    if isinstance(yaml_node, (MappingNode, SequenceNode)):
        nodes, last_token_pos = extract_semantic_structure_info(
            anchor_abs_start=anchor_abs_start,
            last_token_abs_start=last_token_abs_start,
            yaml_node=yaml_node,
            doc=doc,
            semantic_type=semantic_type,
        )

        if not (clean_semantic_type.local_key_fn or clean_semantic_type.index_key_fn):
            return nodes, last_token_pos

        if clean_semantic_type.local_key_fn:
            local_key = clean_semantic_type.local_key_fn(value=yaml_node.value)
        else:
            local_key = None

        if clean_semantic_type.index_key_fn:
            index_key = clean_semantic_type.index_key_fn(value=yaml_node.value)
        else:
            index_key = None

        block = [
            SemanticBlock(
                local_key=local_key,
                index_key=index_key,
                anchor_rel=Range(
                    start=_compute_rel_position(
                        line=yaml_node.start_mark.line,
                        character=yaml_node.start_mark.column,
                        relative_to=anchor_abs_start,
                    ),
                    end=_compute_rel_position(
                        line=yaml_node.end_mark.line,
                        character=yaml_node.end_mark.column,
                        relative_to=anchor_abs_start,
                    ),
                ),
                children=nodes,
            )
        ]
        return block, last_token_pos

    if clean_semantic_type.type:
        node_type = clean_semantic_type.type
    else:
        tag_kind = yaml_node.tag.rsplit(":", 1)[-1]
        if tag_kind in {"int", "float", "bool"}:
            node_type = "number"
        else:
            node_type = "string"

    if node_type == "string" and yaml_node.value.startswith("="):
        return _extract_cel_semantic_info(
            anchor_abs_start=anchor_abs_start,
            last_token_abs_start=last_token_abs_start,
            yaml_node=yaml_node,
            doc=doc,
            modifier=clean_semantic_type.modifier,
            diagnostic=diagnostic,
        )

    clean_semantic_type.type = node_type

    return _extract_scalar_semantic_info(
        anchor_abs_start=anchor_abs_start,
        last_token_abs_start=last_token_abs_start,
        yaml_node=yaml_node,
        doc=doc,
        semantic_type=clean_semantic_type,
        diagnostic=diagnostic,
    )


def _compute_rel_position(line: int, character: int, relative_to: Position) -> Position:
    rel_to_line = relative_to.line
    rel_to_offset = relative_to.character

    rel_line = line - rel_to_line
    return Position(
        line=rel_line,
        character=character if rel_line > 0 else (character - rel_to_offset),
    )



from typing import NamedTuple, Sequence
import os

os.environ["KOREO_DEV_TOOLING"] = "true"

from lsprotocol import types

from koreo import cache
from koreo.function_test.structure import FunctionTest

from koreo_tooling import function_test
from koreo_tooling.indexing import compute_abs_range
from koreo_tooling.indexing.semantics import SemanticAnchor, SemanticBlock, SemanticNode

from koreo_tooling.langserver.rangers import block_range_extract


class RunResults(NamedTuple):
    results: dict[str, function_test.TestResults] | None = None
    logs: list[types.LogMessageParams] | None = None
    diagnostics: list[types.Diagnostic] | None = None


async def run_function_tests(
    tests_to_run: set[str],
    functions_to_test: dict[type, Sequence[str]],
    test_range_map: dict[str, types.Range],
) -> RunResults:
    if not (tests_to_run or functions_to_test):
        return RunResults()

    test_results, logs = await function_test.run_function_tests(
        tests_to_run=tests_to_run,
        functions_to_test=functions_to_test,
    )

    if not test_results:
        return RunResults(logs=logs)

    return RunResults(
        results=test_results,
        logs=logs,
        diagnostics=_process_results(
            tests_to_run=tests_to_run,
            test_results=test_results,
            test_range_map=test_range_map,
        ),
    )


def _process_results(
    tests_to_run: set[str],
    test_results: dict[str, function_test.TestResults],
    test_range_map: dict[str, types.Range],
) -> list[types.Diagnostic]:
    test_diagnostics: list[types.Diagnostic] = []

    for test_key, result in test_results.items():
        if result.success:
            continue

        # TODO: Add support for reporting within Functions
        if test_key not in tests_to_run:
            continue

        cached_resource = cache.get_resource_system_data_from_cache(
            resource_class=FunctionTest, cache_key=test_key
        )
        if not (
            cached_resource and cached_resource.resource and cached_resource.system_data
        ):
            continue

        # This is the test's location within the file.
        anchor = cached_resource.system_data.get("anchor")
        if not anchor:
            continue

        if result.messages:
            test_diagnostics.append(
                types.Diagnostic(
                    message=f"Failures: {'; '.join(result.messages)}",
                    severity=types.DiagnosticSeverity.Error,
                    range=test_range_map[test_key],
                )
            )

        test_spec_block = block_range_extract(
            search_key="spec",
            search_nodes=anchor.children,
            anchor=anchor,
        )
        match test_spec_block:
            case None:
                # Could not find the FunctionTest's spec. Hopefully they're
                # typing it right now.
                continue
            case list(block_diagnostics):
                # There was some type of structural issue that resulted in the
                # spec block lookup returning errors. The diagnostics should
                # explain the issue.
                test_diagnostics.extend(block_diagnostics)
                continue

        if result.input_mismatches:
            test_diagnostics.extend(
                _process_input_errors(
                    input_mismatches=result.input_mismatches,
                    test_spec_block=test_spec_block,
                    anchor=anchor,
                )
            )

        if result.resource_field_errors:
            test_diagnostics.extend(
                _process_resource_errors(
                    field_errors=result.resource_field_errors,
                    test_spec_block=test_spec_block,
                    anchor=anchor,
                )
            )

        if result.outcome_fields_errors:
            test_diagnostics.extend(
                _process_outcome_errors(
                    field_errors=result.outcome_fields_errors,
                    test_spec_block=test_spec_block,
                    anchor=anchor,
                )
            )

    return test_diagnostics


def _process_input_errors(
    input_mismatches: Sequence[function_test.FieldMismatchResult],
    test_spec_block: SemanticAnchor | SemanticBlock | SemanticNode,
    anchor: SemanticAnchor,
) -> list[types.Diagnostic]:
    inputs_block = block_range_extract(
        search_key="inputs",
        search_nodes=test_spec_block.children,
        anchor=anchor,
    )
    match inputs_block:
        case None:
            return [
                types.Diagnostic(
                    message="Input expected, but none provided.",
                    severity=types.DiagnosticSeverity.Warning,
                    range=compute_abs_range(test_spec_block, anchor),
                )
            ]
        case list(block_diagnostics):
            return block_diagnostics

    input_values_block = block_range_extract(
        search_key="input-values",
        search_nodes=inputs_block.children,
        anchor=anchor,
    )
    match input_values_block:
        case None:
            return []
        case list(block_diagnostics):
            return block_diagnostics

    diagnostics: list[types.Diagnostic] = []

    for mismatch in input_mismatches:
        if not mismatch.expected and mismatch.actual:
            input_block = block_range_extract(
                search_key=f"input:{mismatch.field.split(".", 1)[-1]}",
                search_nodes=input_values_block.children,
                anchor=anchor,
            )
            match input_block:
                case None:
                    continue
                case list(block_diagnostics):
                    diagnostics.extend(block_diagnostics)
                    continue

            diagnostics.append(
                types.Diagnostic(
                    message=f"Input ('{mismatch.field}') not expected.",
                    severity=types.DiagnosticSeverity.Warning,
                    range=compute_abs_range(input_block, anchor),
                )
            )

        elif mismatch.expected and not mismatch.actual:
            diagnostics.append(
                types.Diagnostic(
                    message=f"Missing input ('{mismatch.field}').",
                    severity=types.DiagnosticSeverity.Error,
                    range=compute_abs_range(inputs_block, anchor),
                )
            )

    return diagnostics


def _process_resource_errors(
    field_errors: Sequence[function_test.CompareResult],
    test_spec_block: SemanticAnchor | SemanticBlock | SemanticNode,
    anchor: SemanticAnchor,
) -> list[types.Diagnostic]:
    resource_block = block_range_extract(
        search_key="expected_resource",
        search_nodes=test_spec_block.children,
        anchor=anchor,
    )
    match resource_block:
        case None:
            return [
                types.Diagnostic(
                    message="Resource unexpectedly modified.",
                    severity=types.DiagnosticSeverity.Error,
                    range=compute_abs_range(test_spec_block, anchor),
                )
            ]
        case list(block_diagnostics):
            return block_diagnostics

    diagnostics: list[types.Diagnostic] = []

    for mismatch in field_errors:
        mismatch_block = block_range_extract(
            search_key=mismatch.field,
            search_nodes=resource_block.children,
            anchor=anchor,
        )
        match mismatch_block:
            case None:
                diagnostics.append(
                    types.Diagnostic(
                        message=f"Missing value for '{mismatch.field}'",
                        severity=types.DiagnosticSeverity.Error,
                        range=compute_abs_range(resource_block, anchor),
                    )
                )
                continue
            case list(block_diagnostics):
                diagnostics.extend(block_diagnostics)
                continue

        diagnostics.append(
            types.Diagnostic(
                message=f"Actual: '{mismatch.actual}'",
                severity=types.DiagnosticSeverity.Error,
                range=compute_abs_range(mismatch_block, anchor),
            )
        )

    return diagnostics


def _process_outcome_errors(
    field_errors: Sequence[function_test.CompareResult],
    test_spec_block: SemanticAnchor | SemanticBlock | SemanticNode,
    anchor: SemanticAnchor,
) -> list[types.Diagnostic]:
    outcome_block = block_range_extract(
        search_key="expected_return",
        search_nodes=test_spec_block.children,
        anchor=anchor,
    )
    match outcome_block:
        case None:
            return [
                types.Diagnostic(
                    message="Outcome unexpectedly reached.",
                    severity=types.DiagnosticSeverity.Error,
                    range=compute_abs_range(test_spec_block, anchor),
                )
            ]
        case list(block_diagnostics):
            return block_diagnostics

    diagnostics: list[types.Diagnostic] = []

    for mismatch in field_errors:
        mismatch_block = block_range_extract(
            search_key=mismatch.field,
            search_nodes=outcome_block.children,
            anchor=anchor,
        )
        match mismatch_block:
            case None:
                diagnostics.append(
                    types.Diagnostic(
                        message=f"Missing value for '{mismatch.field}'",
                        severity=types.DiagnosticSeverity.Error,
                        range=compute_abs_range(outcome_block, anchor),
                    )
                )
                continue
            case list(block_diagnostics):
                diagnostics.extend(block_diagnostics)
                continue

        diagnostics.append(
            types.Diagnostic(
                message=f"Actual: '{mismatch.actual}'",
                severity=types.DiagnosticSeverity.Error,
                range=compute_abs_range(mismatch_block, anchor),
            )
        )

    return diagnostics



from typing import NamedTuple, Sequence
import copy

import yaml

from lsprotocol import types

from koreo import cache
from koreo.function_test.structure import FunctionTest

from koreo_tooling import constants
from koreo_tooling.function_test import FieldMismatchResult, TestResults
from koreo_tooling.indexing.semantics import (
    SemanticAnchor,
    SemanticBlock,
    SemanticNode,
    compute_abs_range,
    compute_abs_position,
)
from koreo_tooling.langserver.rangers import block_range_extract


class LensResult(NamedTuple):
    lens: Sequence[types.CodeLens] | None = None
    logs: Sequence[types.LogMessageParams] | None = None


class EditResult(NamedTuple):
    edits: Sequence[types.TextEdit] | None = None
    logs: Sequence[types.LogMessageParams] | None = None


def handle_lens(
    uri: str, doc_uri: str, doc_version: int, test_results: dict[str, TestResults]
):
    if not test_results:
        return LensResult()

    logs = []

    lens = []
    for test_name, test_result in test_results.items():
        if test_result.success:
            continue

        cached = cache.get_resource_system_data_from_cache(
            resource_class=FunctionTest, cache_key=test_name
        )

        if not (cached and cached.resource and cached.system_data):
            continue

        if cached.system_data.get("uri") != uri:
            continue

        test_anchor = cached.system_data.get("anchor")
        if not test_anchor:
            continue

        if test_result.input_mismatches:
            lens_result = _generate_inputs_lens(
                doc_uri=doc_uri,
                doc_version=doc_version,
                test_name=test_name,
                input_mismatches=test_result.input_mismatches,
                test_anchor=test_anchor,
            )
            if lens_result.lens:
                lens.extend(lens_result.lens)

            if lens_result.logs:
                logs.extend(lens_result.logs)

        if test_result.actual_resource and test_result.missing_test_assertion:
            lens_result = _generate_current_resource_lens(
                doc_uri=doc_uri,
                doc_version=doc_version,
                test_name=test_name,
                test_anchor=test_anchor,
            )
            if lens_result.lens:
                lens.extend(lens_result.lens)

            if lens_result.logs:
                logs.extend(lens_result.logs)

        if test_result.resource_field_errors or test_result.missing_test_assertion:
            lens_result = _generate_resource_lens(
                doc_uri=doc_uri,
                doc_version=doc_version,
                test_name=test_name,
                test_anchor=test_anchor,
            )
            if lens_result.lens:
                lens.extend(lens_result.lens)

            if lens_result.logs:
                logs.extend(lens_result.logs)

        if test_result.outcome_fields_errors or test_result.missing_test_assertion:
            lens_result = _generate_return_value_lens(
                doc_uri=doc_uri,
                doc_version=doc_version,
                test_name=test_name,
                test_anchor=test_anchor,
            )
            if lens_result.lens:
                lens.extend(lens_result.lens)

            if lens_result.logs:
                logs.extend(lens_result.logs)

    return LensResult(lens=lens, logs=logs)


def _generate_inputs_lens(
    doc_uri: str,
    doc_version: int,
    test_name: str,
    input_mismatches: list[FieldMismatchResult],
    test_anchor: SemanticAnchor,
) -> LensResult:
    if not any(mismatch.field.startswith("inputs.") for mismatch in input_mismatches):
        return LensResult()

    inputs_block = block_range_extract(
        search_key="inputs",
        search_nodes=test_anchor.children,
        anchor=test_anchor,
    )
    match inputs_block:
        case None:
            return LensResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"Test ({test_name}) missing inputs block",
                    )
                ]
            )
        case list(_):
            return LensResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"Test ({test_name}) has duplicate inputs blocks",
                    )
                ]
            )

    return LensResult(
        lens=[
            types.CodeLens(
                range=compute_abs_range(inputs_block, test_anchor),
                command=types.Command(
                    title="Autocorrect Inputs",
                    command="codeLens.completeInputs",
                    arguments=[
                        {
                            "uri": doc_uri,
                            "version": doc_version,
                            "test_name": test_name,
                        }
                    ],
                ),
            )
        ]
    )


def _generate_resource_lens(
    doc_uri: str,
    doc_version: int,
    test_name: str,
    test_anchor: SemanticAnchor,
) -> LensResult:
    expected_resource_block = block_range_extract(
        search_key="expected_resource",
        search_nodes=test_anchor.children,
        anchor=test_anchor,
    )
    match expected_resource_block:
        case None:
            return LensResult()
        case list(_):
            return LensResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"FunctionTest ({test_name}) has multiple expectResource blocks",
                    )
                ]
            )

    return LensResult(
        lens=[
            types.CodeLens(
                range=compute_abs_range(expected_resource_block, test_anchor),
                command=types.Command(
                    title="Autocorrect expected resource",
                    command="codeLens.completeExpectedResource",
                    arguments=[
                        {
                            "uri": doc_uri,
                            "version": doc_version,
                            "test_name": test_name,
                        }
                    ],
                ),
            )
        ]
    )


def _generate_current_resource_lens(
    doc_uri: str,
    doc_version: int,
    test_name: str,
    test_anchor: SemanticAnchor,
) -> LensResult:
    current_resource_block = block_range_extract(
        search_key="current_resource",
        search_nodes=test_anchor.children,
        anchor=test_anchor,
    )
    match current_resource_block:
        case None:
            return LensResult()
        case list(_):
            return LensResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"FunctionTest ({test_name}) has multiple currentResource blocks",
                    )
                ]
            )

    return LensResult(
        lens=[
            types.CodeLens(
                range=compute_abs_range(current_resource_block, test_anchor),
                command=types.Command(
                    title="Autofill current resource",
                    command="codeLens.completeCurrentResource",
                    arguments=[
                        {
                            "uri": doc_uri,
                            "version": doc_version,
                            "test_name": test_name,
                        }
                    ],
                ),
            )
        ]
    )


def _generate_return_value_lens(
    doc_uri: str,
    doc_version: int,
    test_name: str,
    test_anchor: SemanticAnchor,
) -> LensResult:
    return_value_block = block_range_extract(
        search_key="expected_return",
        search_nodes=test_anchor.children,
        anchor=test_anchor,
    )
    match return_value_block:
        case None:
            return LensResult()
        case list(_):
            return LensResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"Test ({test_name}) has multiple expectReturn blocks",
                    )
                ]
            )

    return LensResult(
        lens=[
            types.CodeLens(
                range=compute_abs_range(return_value_block, test_anchor),
                command=types.Command(
                    title="Autocorrect expected return value",
                    command="codeLens.completeExpectedReturnValue",
                    arguments=[
                        {
                            "uri": doc_uri,
                            "version": doc_version,
                            "test_name": test_name,
                        }
                    ],
                ),
            )
        ]
    )


def _code_lens_inputs_action(test_name: str, test_result: TestResults):
    if not test_result.input_mismatches:
        return EditResult()

    cached = cache.get_resource_system_data_from_cache(
        resource_class=FunctionTest, cache_key=test_name
    )
    if not (cached and cached.resource and cached.system_data):
        return EditResult(
            logs=[
                types.LogMessageParams(
                    type=types.MessageType.Debug,
                    message=f"test ({test_name}) isn't cached right",
                )
            ]
        )

    test_anchor = cached.system_data.get("anchor")
    if not test_anchor:
        return EditResult(
            logs=[
                types.LogMessageParams(
                    type=types.MessageType.Debug,
                    message=f"test ({test_name}) cache missing anchor",
                )
            ]
        )

    inputs_block = block_range_extract(
        search_key="inputs",
        search_nodes=test_anchor.children,
        anchor=test_anchor,
    )
    match inputs_block:
        case None:
            return EditResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"test ({test_name}) missing inputs block",
                    )
                ]
            )
        case list(_):
            return EditResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"test ({test_name}) duplicate inputs block",
                    )
                ]
            )

    inputs_value_block = block_range_extract(
        search_key="input-values",
        search_nodes=inputs_block.children,
        anchor=test_anchor,
    )
    match inputs_value_block:
        case None | list(_):
            return EditResult()

    # We're going to mutate this object, so make a copy.
    spec_inputs = copy.deepcopy(cached.spec.get("inputs"))

    logs: list[types.LogMessageParams] = []

    input_fields = (
        (input_match.group("name"), mismatch)
        for input_match, mismatch in (
            (constants.INPUT_NAME_PATTERN.match(mismatch.field), mismatch)
            for mismatch in test_result.input_mismatches
        )
        if input_match
    )

    place_holder = "TODO"

    if not spec_inputs:
        spec_inputs = {
            input_name: place_holder
            for (input_name, mismatch) in input_fields
            if mismatch.expected
        }
    else:
        for input_name, mismatch in input_fields:
            if mismatch.actual and not mismatch.expected:
                del spec_inputs[input_name]

            if not mismatch.actual and mismatch.expected:
                spec_inputs[input_name] = place_holder

    offset, edit_range = _compute_edit_range(
        label=inputs_block, value_block=inputs_value_block, anchor=test_anchor
    )

    indent = (offset + 2) * " "
    formated_inputs = f"\n{"\n".join(
        f"{indent}{line}"
        for line in yaml.dump(spec_inputs).splitlines()
    )}\n\n"

    return EditResult(
        edits=(types.TextEdit(new_text=formated_inputs, range=edit_range),),
        logs=logs,
    )


def _code_lens_current_resource_action(test_name: str, test_result: TestResults):
    if not (test_result.actual_resource and test_result.missing_test_assertion):
        return EditResult()

    return _code_lens_replace_value_block_action(
        test_name=test_name,
        label_block_key="current_resource",
        value_block_key="current_value",
        new_value=test_result.actual_resource,
    )


def _code_lens_resource_action(test_name: str, test_result: TestResults):
    if not (test_result.resource_field_errors or test_result.missing_test_assertion):
        return EditResult()

    return _code_lens_replace_value_block_action(
        test_name=test_name,
        label_block_key="expected_resource",
        value_block_key="expected_value",
        new_value=test_result.actual_resource,
    )


def _code_lens_return_value_action(test_name: str, test_result: TestResults):
    if not (test_result.outcome_fields_errors or test_result.missing_test_assertion):
        return EditResult()

    return _code_lens_replace_value_block_action(
        test_name=test_name,
        label_block_key="expected_return",
        value_block_key="expected_value",
        new_value=test_result.actual_return,
    )


def _code_lens_replace_value_block_action(
    test_name: str,
    label_block_key: str,
    value_block_key: str,
    new_value: dict | None,
):
    if new_value is None:
        return EditResult(
            logs=[
                types.LogMessageParams(
                    type=types.MessageType.Debug,
                    message=f"FunctionTest ({test_name}) can not auto-complete {label_block_key}",
                )
            ]
        )

    cached = cache.get_resource_system_data_from_cache(
        resource_class=FunctionTest, cache_key=test_name
    )
    if not (cached and cached.resource and cached.system_data):
        return EditResult(
            logs=[
                types.LogMessageParams(
                    type=types.MessageType.Debug,
                    message=f"FunctionTest ({test_name}) isn't cached, or cache is corrupt",
                )
            ]
        )

    test_anchor = cached.system_data.get("anchor")
    if not test_anchor:
        return EditResult(
            logs=[
                types.LogMessageParams(
                    type=types.MessageType.Debug,
                    message=f"FunctionTest ({test_name}) cache corrupt, missing anchor",
                )
            ]
        )

    label_block = block_range_extract(
        search_key=label_block_key,
        search_nodes=test_anchor.children,
        anchor=test_anchor,
    )
    match label_block:
        case None:
            return EditResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"FunctionTest ({test_name}) missing {label_block_key} block",
                    )
                ]
            )
        case list(_):
            return EditResult(
                logs=[
                    types.LogMessageParams(
                        type=types.MessageType.Debug,
                        message=f"FunctionTest ({test_name}) duplicate {label_block_key} block",
                    )
                ]
            )

    value_block = block_range_extract(
        search_key=value_block_key,
        search_nodes=label_block.children,
        anchor=test_anchor,
    )
    match value_block:
        case None | list(_):
            return EditResult()

    logs: list[types.LogMessageParams] = []

    offset, edit_range = _compute_edit_range(
        label=label_block, value_block=value_block, anchor=test_anchor
    )

    indent = (offset + 2) * " "
    formated = f"\n{"\n".join(
        f"{indent}{line}"
        for line in yaml.dump(new_value, width=10000).splitlines()
    )}\n\n"

    return EditResult(
        edits=(types.TextEdit(new_text=formated, range=edit_range),),
        logs=logs,
    )


def _compute_edit_range(
    label: SemanticAnchor | SemanticBlock | SemanticNode,
    value_block: SemanticAnchor | SemanticBlock | SemanticNode,
    anchor: SemanticAnchor,
) -> tuple[int, types.Range]:
    match label:
        case SemanticAnchor(abs_position=abs_position):
            # TODO: This would be wrong. Probably should be first-child-start?
            edit_start_pos = abs_position
            offset = abs_position.character
        case SemanticBlock(anchor_rel=anchor_rel):
            # TODO: This would be wrong. Probably should be first-child-start?
            edit_start_pos = compute_abs_position(
                rel_position=anchor_rel.start, abs_position=anchor.abs_position
            )
            offset = anchor_rel.start.character
        case SemanticNode(anchor_rel=anchor_rel, length=length):
            edit_start_pos = compute_abs_position(
                rel_position=anchor_rel,
                abs_position=anchor.abs_position,
                length=length + 1,
            )
            offset = anchor_rel.character

    match value_block:
        case SemanticAnchor(abs_position=abs_position):
            # TODO: This should really never be an anchor node, but if it is,
            # this should probably be the end-position of the deepest last
            # child.
            edit_end_pos = abs_position
        case SemanticBlock(anchor_rel=anchor_rel):
            end_line = anchor_rel.end.line + anchor.abs_position.line
            character = 0
            if end_line == edit_start_pos.line:
                character = anchor_rel.end.character

            edit_end_pos = types.Position(
                line=end_line,
                character=character,
            )
        case SemanticNode(anchor_rel=anchor_rel, length=length):
            edit_end_pos = compute_abs_position(
                rel_position=anchor_rel,
                abs_position=anchor.abs_position,
                length=length,
            )

    return offset, types.Range(start=edit_start_pos, end=edit_end_pos)


LENS_COMMANDS = (
    ("codeLens.completeInputs", _code_lens_inputs_action),
    ("codeLens.completeCurrentResource", _code_lens_current_resource_action),
    ("codeLens.completeExpectedResource", _code_lens_resource_action),
    ("codeLens.completeExpectedReturnValue", _code_lens_return_value_action),
)



from collections import defaultdict
from pathlib import Path
import pathlib
from typing import Callable, NamedTuple, Sequence
import asyncio
import time

import logging

logger = logging.getLogger("koreo.ls")

from lsprotocol import types
from pygls.lsp.server import LanguageServer
from koreo import schema

KOREO_LSP_NAME = "koreo-ls"
KOREO_LSP_VERSION = "v1beta1"
CRD_ROOT = pathlib.Path(__file__).parent.joinpath("crd")

server = LanguageServer(KOREO_LSP_NAME, KOREO_LSP_VERSION)

from koreo import cache
from koreo import registry
from koreo.function_test.structure import FunctionTest
from koreo.resource_function.structure import ResourceFunction
from koreo.resource_template.structure import ResourceTemplate
from koreo.result import is_unwrapped_ok
from koreo.value_function.structure import ValueFunction
from koreo.workflow.structure import Workflow

from koreo_tooling import constants
from koreo_tooling.function_test import TestResults
from koreo_tooling.indexing import TokenModifiers, TokenTypes, SemanticAnchor
from koreo_tooling.indexing.semantics import generate_local_range_index
from koreo_tooling.langserver.codelens import LENS_COMMANDS, EditResult, handle_lens
from koreo_tooling.langserver.fileprocessor import (
    ProccessResults,
    SemanticRangeIndex,
    process_file,
)
from koreo_tooling.langserver.function_test import run_function_tests
from koreo_tooling.langserver.hover import handle_hover
from koreo_tooling.langserver.orchestrator import handle_file, shutdown_handlers
from koreo_tooling.langserver.workflow import process_workflows


@server.feature(types.TEXT_DOCUMENT_COMPLETION)
async def completions(params: types.CompletionParams):
    server.window_log_message(
        params=types.LogMessageParams(
            type=types.MessageType.Debug, message=f"completion params: {params}"
        )
    )

    # TODO: Add awareness of the context to surface the correct completions.
    # Use _lookup_current_line_info to find the context, then offer suitable
    # suggestion, probably similar to hover.

    items = []

    for cache_type, cached in cache.__CACHE.items():
        for resource_key, _ in cached.items():
            items.append(
                types.CompletionItem(
                    label=resource_key,
                    label_details=types.CompletionItemLabelDetails(
                        detail=f" {cache_type}"
                    ),
                )
            )

    # TODO: If we're confident about where they are, set this as is_incomplete=False
    return types.CompletionList(is_incomplete=True, items=items)


@server.feature(types.TEXT_DOCUMENT_HOVER)
def hover(params: types.HoverParams):
    doc = server.workspace.get_text_document(params.text_document.uri)

    resource_info = _lookup_current_line_info(uri=doc.uri, line=params.position.line)
    if not resource_info.index_match:
        return None

    hover_result = handle_hover(
        resource_key=resource_info.index_match.key,
        resource_key_range=resource_info.index_match.range,
        local_resource=resource_info.local_match,
        test_results=__TEST_RESULTS[doc.uri],
    )

    if not hover_result:
        return None

    if hover_result.logs:
        for log in hover_result.logs:
            server.window_log_message(params=log)

    return hover_result.hover


@server.feature(types.TEXT_DOCUMENT_INLAY_HINT)
def inlay_hints(params: types.InlayHintParams):
    doc = server.workspace.get_text_document(params.text_document.uri)

    test_results = __TEST_RESULTS[doc.uri]
    if not test_results:
        return []

    start_line = params.range.start.line
    end_line = params.range.end.line

    visible_tests = []
    for uri, maybe_resource, maybe_range, _ in __SEMANTIC_RANGE_INDEX:
        if uri != doc.uri:
            continue

        if not isinstance(maybe_range, types.Range):
            # TODO: Get anchors setting a range
            continue

        if not (start_line <= maybe_range.start.line <= end_line):
            continue

        match = constants.FUNCTION_TEST_NAME.match(maybe_resource)
        if not match:
            continue

        visible_tests.append((match.group("name"), maybe_range))

    if not visible_tests:
        return []

    inlays = []
    for test_name, name_range in visible_tests:
        result = __TEST_RESULTS[doc.uri].get(test_name)

        if not result:
            inlay = "Not Ran"
        elif result.success:
            inlay = "Success"
        else:
            inlay = "Error"

        char = len(doc.lines[name_range.end.line])

        inlays.append(
            types.InlayHint(
                label=inlay,
                kind=types.InlayHintKind.Type,
                padding_left=True,
                padding_right=True,
                position=types.Position(line=name_range.end.line, character=char),
            )
        )

    return inlays


@server.feature(types.INITIALIZE)
async def initialize(params):
    await _process_workspace_directories()


@server.feature(types.TEXT_DOCUMENT_CODE_LENS)
def code_lens(params: types.CodeLensParams):
    """Return a list of code lens to insert into the given document.

    This method will read the whole document and identify each sum in the document and
    tell the language client to insert a code lens at each location.
    """
    doc = server.workspace.get_text_document(params.text_document.uri)

    test_results = __TEST_RESULTS[doc.uri]
    if not test_results:
        return []

    lens_result = handle_lens(
        doc_uri=params.text_document.uri,
        uri=doc.uri,
        doc_version=doc.version if doc.version else -1,
        test_results=test_results,
    )

    if lens_result.logs:
        for log in lens_result.logs:
            server.window_log_message(params=log)

    return lens_result.lens


def _lens_action_wrapper(fn: Callable[[str, TestResults], EditResult]):
    async def wrapper(args):
        if not args:
            return

        edit_args = args[0]

        doc_uri = edit_args["uri"]
        doc_version = edit_args["version"]
        test_name = edit_args["test_name"]

        doc = server.workspace.get_text_document(doc_uri)
        if doc.version != doc_version:
            return

        doc_uri = doc.uri

        test_result = __TEST_RESULTS[doc_uri][test_name]

        edit_result = fn(test_name, test_result)
        if edit_result.logs:
            for log in edit_result.logs:
                server.window_log_message(params=log)

        if edit_result.edits:
            # Apply the edit.
            server.workspace_apply_edit(
                types.ApplyWorkspaceEditParams(
                    edit=types.WorkspaceEdit(
                        document_changes=[
                            types.TextDocumentEdit(
                                text_document=types.OptionalVersionedTextDocumentIdentifier(
                                    uri=doc_uri,
                                    version=doc.version,
                                ),
                                edits=edit_result.edits,
                            )
                        ]
                    ),
                )
            )

    return wrapper


# Register Code Lens Commands
def _register_lens_commands(lens_commands):
    for command_name, command_fn in lens_commands:
        server.command(command_name=command_name)(_lens_action_wrapper(command_fn))


_register_lens_commands(LENS_COMMANDS)


@server.feature(types.WORKSPACE_DID_CHANGE_CONFIGURATION)
async def change_workspace_config(params):
    # TODO: This should probably load a config file that enumerates
    # namespaces or something.

    await _process_workspace_directories()


async def _process_workspace_directories():
    suffixes = ("yaml", "yml", "koreo")

    for folder_key in server.workspace.folders:
        path = Path(server.workspace.get_text_document(folder_key).path)

        for suffix in suffixes:
            for match in path.rglob(f"*.{suffix}"):

                await handle_file(
                    file_uri=f"file://{match}",
                    monotime=time.monotonic(),
                    file_processor=_parse_file,
                )


@server.feature(types.TEXT_DOCUMENT_DID_OPEN)
async def open_processor(params):

    await handle_file(
        file_uri=params.text_document.uri,
        monotime=time.monotonic(),
        file_processor=_parse_file,
    )


@server.feature(types.TEXT_DOCUMENT_DID_CHANGE)
async def change_processor(params):

    await handle_file(
        file_uri=params.text_document.uri,
        monotime=time.monotonic(),
        file_processor=_parse_file,
    )


class ResourceMatch(NamedTuple):
    key: str
    range: types.Range


class CurrentLineInfo(NamedTuple):
    index_match: ResourceMatch | None = None
    local_match: ResourceMatch | None = None


def _lookup_current_line_info(uri: str, line: int) -> CurrentLineInfo:
    possible_match: tuple[str, types.Range] | None = None
    for maybe_uri, maybe_key, maybe_range, _ in __SEMANTIC_RANGE_INDEX:
        if maybe_uri != uri:
            continue

        if not isinstance(maybe_range, types.Range):
            continue

        if maybe_range.start.line == line:
            return CurrentLineInfo(
                index_match=ResourceMatch(key=maybe_key, range=maybe_range)
            )

        if maybe_range.start.line <= line <= maybe_range.end.line:
            possible_match = ResourceMatch(key=maybe_key, range=maybe_range)

    if not possible_match:
        return CurrentLineInfo()

    if match := constants.WORKFLOW_ANCHOR.match(possible_match.key):
        cached = cache.get_resource_system_data_from_cache(
            resource_class=Workflow, cache_key=match.group("name")
        )
    elif match := constants.FUNCTION_TEST_ANCHOR.match(possible_match.key):
        cached = cache.get_resource_system_data_from_cache(
            resource_class=FunctionTest, cache_key=match.group("name")
        )
    elif match := constants.RESOURCE_FUNCTION_ANCHOR.match(possible_match.key):
        cached = cache.get_resource_system_data_from_cache(
            resource_class=ResourceFunction, cache_key=match.group("name")
        )
    elif match := constants.VALUE_FUNCTION_ANCHOR.match(possible_match.key):
        cached = cache.get_resource_system_data_from_cache(
            resource_class=ValueFunction, cache_key=match.group("name")
        )
    else:
        return CurrentLineInfo()

    if not cached or not cached.system_data or "anchor" not in cached.system_data:
        return CurrentLineInfo()

    anchor: SemanticAnchor = cached.system_data.get("anchor")
    anchor_index = generate_local_range_index(nodes=anchor.children, anchor=anchor)

    for maybe_key, maybe_range in anchor_index:
        if not isinstance(maybe_range, types.Range):
            continue

        if maybe_range.start.line == line:
            return CurrentLineInfo(
                index_match=possible_match,
                local_match=ResourceMatch(key=maybe_key, range=maybe_range),
            )

    return CurrentLineInfo()


@server.feature(types.TEXT_DOCUMENT_DEFINITION)
async def goto_definitiion(params: types.DefinitionParams):
    doc = server.workspace.get_text_document(params.text_document.uri)

    resource_info = _lookup_current_line_info(uri=doc.uri, line=params.position.line)
    if not resource_info.index_match:
        return []

    resource_key_root = resource_info.index_match.key.rsplit(":", 1)[0]
    search_key = f"{resource_key_root}:def"

    definitions = []
    for uri, maybe_key, maybe_range, _ in __SEMANTIC_RANGE_INDEX:
        if not isinstance(maybe_range, types.Range):
            # TODO: Get anchors setting a range
            continue

        if maybe_key != search_key:
            continue

        definitions.append(types.Location(uri=uri, range=maybe_range))

    return definitions


@server.feature(types.TEXT_DOCUMENT_REFERENCES)
async def goto_reference(params: types.ReferenceParams):
    doc = server.workspace.get_text_document(params.text_document.uri)

    resource_info = _lookup_current_line_info(uri=doc.uri, line=params.position.line)
    if not resource_info.index_match:
        return []

    resource_key_root = resource_info.index_match.key.rsplit(":", 1)[0]
    reference_key = f"{resource_key_root}:ref"
    definition_key = f"{resource_key_root}:def"

    references: list[types.Location] = []
    definitions: list[types.Location] = []
    for uri, maybe_key, maybe_range, _ in __SEMANTIC_RANGE_INDEX:
        if not isinstance(maybe_range, types.Range):
            # TODO: Get anchors setting a range
            continue

        if maybe_key == reference_key:
            references.append(types.Location(uri=uri, range=maybe_range))

        if maybe_key == definition_key:
            definitions.append(types.Location(uri=uri, range=maybe_range))

    return references + definitions


@server.feature(
    types.TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
    types.SemanticTokensLegend(token_types=TokenTypes, token_modifiers=TokenModifiers),
)
async def semantic_tokens_full(params: types.ReferenceParams):
    doc = server.workspace.get_text_document(params.text_document.uri)
    if not doc.uri in __SEMANTIC_TOKEN_INDEX:
        # Once processing has finished, a refresh will be issued.
        await handle_file(
            file_uri=params.text_document.uri,
            monotime=time.monotonic(),
            file_processor=_parse_file,
        )
        return

    tokens = __SEMANTIC_TOKEN_INDEX[doc.uri]
    return types.SemanticTokens(data=tokens)


@server.feature(types.SHUTDOWN)
async def shutdown(*_, **__):
    with open("/Users/bobert/tmp/koreo-ls.log", "w") as outfile:
        outfile.write("shutting down")
        await shutdown_handlers()
        outfile.write("shut down complete")


class ParseNotification:
    pass


async def _parse_file(doc_uri: str):
    doc = server.workspace.get_text_document(doc_uri)

    file_analyzer_resource = registry.Resource(resource_type=FileAnalyzer, name=doc_uri)
    parse_notification_resource = registry.Resource(
        resource_type=ParseNotification, name=doc_uri
    )
    if file_analyzer_resource not in _ANALYZERS:
        registry.register(registerer=file_analyzer_resource)
        registry.subscribe(
            subscriber=file_analyzer_resource, resource=parse_notification_resource
        )

        analysis_task = asyncio.create_task(
            _analyzer_manager(file_analyzer_resource), name=doc_uri
        )
        _ANALYZERS[file_analyzer_resource] = analysis_task
        analysis_task.add_done_callback(
            lambda _: _ANALYZERS.__delitem__(file_analyzer_resource)
        )

    try:
        processing_result = await process_file(doc=doc)

        if processing_result.logs:
            for log in processing_result.logs:
                server.window_log_message(params=log)

    except Exception as err:
        return

    __PARSING_RESULT[doc_uri] = (doc.version, processing_result)

    registry.notify_subscribers(
        notifier=parse_notification_resource, event_time=time.monotonic()
    )


async def _run_doc_analysis(doc_uri: str):
    parse_info = __PARSING_RESULT.get(doc_uri)
    if not parse_info:
        return

    doc_version, parsing_result = parse_info

    await _analyze_file(
        doc_uri=doc_uri,
        doc_version=doc_version,
        parsing_result=parsing_result,
    )


async def _analyze_file(
    doc_uri: str,
    doc_version: int | None,
    parsing_result: ProccessResults,
):
    doc = server.workspace.get_text_document(doc_uri)

    if doc_version != doc.version:
        return

    diagnostics: list[types.Diagnostic] = []

    if parsing_result.diagnostics:
        diagnostics.extend(parsing_result.diagnostics)

    diagnostics.extend(
        _process_workflows(
            uri=doc.uri, semantic_range_index=parsing_result.semantic_range_index
        )
    )

    test_diagnostics, test_results = await _run_function_test(
        semantic_range_index=parsing_result.semantic_range_index
    )
    diagnostics.extend(test_diagnostics)

    if parsing_result.semantic_range_index:
        used_resources = _get_used_resources(parsing_result.semantic_range_index)
        current_resource_defs: dict[registry.Resource, str | None] = (
            _get_defined_resources(parsing_result.semantic_range_index)
        )
    else:
        used_resources: list[registry.Resource] = []
        current_resource_defs: dict[registry.Resource, str | None] = {}

    diagnostics.extend(
        _check_defined_resources(
            current_resource_defs.keys(), parsing_result.semantic_range_index
        )
    )

    old_resource_defs: dict[registry.Resource, str | None] = _get_defined_resources(
        [
            range_index
            for range_index in __SEMANTIC_RANGE_INDEX
            if range_index.uri == doc.uri
        ]
    )
    _reset_file_state(doc.uri)

    if parsing_result.semantic_tokens:
        __SEMANTIC_TOKEN_INDEX[doc.uri] = parsing_result.semantic_tokens
    else:
        # Will this break anything?
        __SEMANTIC_TOKEN_INDEX[doc.uri] = []

    if parsing_result.semantic_range_index:
        __SEMANTIC_RANGE_INDEX.extend(parsing_result.semantic_range_index)

    file_analyzer_resource = registry.Resource(resource_type=FileAnalyzer, name=doc_uri)
    used_resources.append(
        registry.Resource(resource_type=ParseNotification, name=doc_uri)
    )
    registry.subscribe_only_to(
        subscriber=file_analyzer_resource, resources=used_resources
    )
    if file_analyzer_resource not in _ANALYZERS:
        analysis_task = asyncio.create_task(
            _analyzer_manager(file_analyzer_resource), name=doc_uri
        )
        _ANALYZERS[file_analyzer_resource] = analysis_task
        analysis_task.add_done_callback(
            lambda _: _ANALYZERS.__delitem__(file_analyzer_resource)
        )

    deleted_resource_defs = set(old_resource_defs.keys()).difference(
        current_resource_defs.keys()
    )
    for deleted_resource in deleted_resource_defs:
        await cache.delete_from_cache(
            resource_class=deleted_resource.resource_type,
            cache_key=deleted_resource.name,
            version=old_resource_defs[deleted_resource],
        )

    __TEST_RESULTS[doc.uri] = test_results

    diagnostics.extend(_check_for_duplicate_resources(uri=doc.uri))

    server.text_document_publish_diagnostics(
        types.PublishDiagnosticsParams(
            uri=doc_uri, version=doc_version, diagnostics=diagnostics
        )
    )


class FileAnalyzer: ...


_ANALYZERS: dict[registry.Resource[FileAnalyzer], asyncio.Task] = {}


async def _analyzer_manager(file_analyzer_resource: registry.Resource):
    queue = registry.register(registerer=file_analyzer_resource)
    last_time = 0
    while True:
        event = await queue.get()
        try:
            match event:
                case registry.Kill():
                    logger.info(f"Killing Analysis Manager ({file_analyzer_resource})")
                    break
                case registry.ResourceEvent(_, event_time):
                    await _run_doc_analysis(
                        doc_uri=file_analyzer_resource.name,
                    )
                    server.workspace_inlay_hint_refresh(None)
                    server.workspace_code_lens_refresh(None)
                    last_time = event_time
                    continue
        finally:
            queue.task_done()


def _get_defined_resources(semantic_range_index: Sequence[SemanticRangeIndex]):
    definitions = (
        (match.group("kind"), match.group("name"), version)
        for match, version in (
            (constants.RESOURCE_DEF.match(range_index.name), range_index.version)
            for range_index in semantic_range_index
        )
        if match
    )

    resources = dict[registry.Resource, str | None]()
    for definition in definitions:
        match definition:
            case ("ValueFunction", name, version):
                resource_key = registry.Resource(resource_type=ValueFunction, name=name)
            case ("ResourceFunction", name, version):
                resource_key = registry.Resource(
                    resource_type=ResourceFunction, name=name
                )
            case ("FunctionTest", name, version):
                resource_key = registry.Resource(resource_type=FunctionTest, name=name)
            case ("Workflow", name, version):
                resource_key = registry.Resource(resource_type=Workflow, name=name)
            case ("ResourceTemplate", name, version):
                resource_key = registry.Resource(
                    resource_type=ResourceTemplate, name=name
                )
            case _:
                continue

        resources[resource_key] = version
    return resources


def _get_used_resources(semantic_range_index: Sequence[SemanticRangeIndex]):
    references = set(
        (match.group("kind"), match.group("name"))
        for match in (
            constants.TOP_LEVEL_RESOURCE.match(range_index.name)
            for range_index in semantic_range_index
        )
        if match
    )

    resources: list[registry.Resource] = []
    for reference in references:
        match reference:
            case ("ResourceFunction", name):
                resources.append(
                    registry.Resource(resource_type=ResourceFunction, name=name)
                )
            case ("ValueFunction", name):
                resources.append(
                    registry.Resource(resource_type=ValueFunction, name=name)
                )
            case ("FunctionTest", name):
                resources.append(
                    registry.Resource(resource_type=FunctionTest, name=name)
                )
            case ("Workflow", name):
                resources.append(registry.Resource(resource_type=Workflow, name=name))
            case ("ResourceTemplate", name):
                resources.append(
                    registry.Resource(resource_type=ResourceTemplate, name=name)
                )
    return resources


def _check_defined_resources(
    resources: Sequence[registry.Resource],
    semantic_range_index: Sequence[SemanticRangeIndex] | None,
) -> Sequence[types.Diagnostic]:
    diagnostics: list[types.Diagnostic] = []

    if not semantic_range_index:
        return diagnostics

    resource_defs = {
        f"{match.group("kind")}:{match.group("name")}": match_range
        for match, match_range in (
            (constants.RESOURCE_DEF.match(range_index.name), range_index.range)
            for range_index in semantic_range_index
        )
        if match
    }

    for resource in resources:
        if resource.resource_type is Workflow:
            continue

        cached = cache.get_resource_from_cache(
            resource_class=resource.resource_type, cache_key=resource.name
        )

        if is_unwrapped_ok(cached):
            continue

        resource_range = resource_defs.get(
            f"{resource.resource_type.__name__}:{resource.name}"
        )
        if not resource_range:
            continue

        diagnostics.append(
            types.Diagnostic(
                message=cached.message,
                severity=types.DiagnosticSeverity.Error,
                range=resource_range,
            )
        )
    return diagnostics


def _process_workflows(
    uri: str, semantic_range_index: Sequence[SemanticRangeIndex] | None
) -> list[types.Diagnostic]:
    if not semantic_range_index:
        return []

    workflows = []
    for _, resource_key, resource_range, _ in semantic_range_index:
        match = constants.WORKFLOW_NAME.match(resource_key)
        if not match:
            continue

        workflows.append((match.group("name"), resource_range))

    return process_workflows(uri=uri, workflows=workflows)


async def _run_function_test(
    semantic_range_index: Sequence[SemanticRangeIndex] | None,
) -> tuple[list[types.Diagnostic], dict[str, TestResults]]:
    """For the given `semantic_range_index`, get all `FunctionTests`. It will
    also accumulate all `ResourceFunction` and `ValueFunction` resources, the
    test runner will lookup all associated `FunctionTests` and run them as
    well. This allows tests and functions in different files, to cause the
    tests are re-run update test results.
    """
    if not semantic_range_index:
        return ([], {})

    test_range_map = {}
    tests_to_run = set[str]()
    resource_functions_to_test = set[str]()
    value_functions_to_test = set[str]()

    for _, resource_key, resource_range, _ in semantic_range_index:
        if match := constants.FUNCTION_TEST_NAME.match(resource_key):
            test_name = match.group("name")
            tests_to_run.add(test_name)
            test_range_map[test_name] = resource_range

        elif match := constants.RESOURCE_FUNCTION_NAME.match(resource_key):
            resource_functions_to_test.add(match.group("name"))

        elif match := constants.VALUE_FUNCTION_NAME.match(resource_key):
            value_functions_to_test.add(match.group("name"))

    if not (tests_to_run or value_functions_to_test or resource_functions_to_test):
        return ([], {})

    functions_to_test: dict[type, Sequence[str]] = {
        ValueFunction: tuple(value_functions_to_test),
        ResourceFunction: tuple(resource_functions_to_test),
    }

    test_results = await run_function_tests(
        tests_to_run=tests_to_run,
        functions_to_test=functions_to_test,
        test_range_map=test_range_map,
    )

    results = test_results.results
    if not results:
        results = {}

    if test_results.logs:
        for log in test_results.logs:
            server.window_log_message(params=log)

    if test_results.diagnostics:
        return (test_results.diagnostics, results)

    return ([], results)


# Globals? Are you evil or just stupid? Yes.
__PARSING_RESULT: dict[str, tuple[int | None, ProccessResults]] = {}
__TEST_RESULTS = defaultdict[str, dict[str, TestResults]](dict)
__SEMANTIC_TOKEN_INDEX: dict[str, Sequence[int]] = {}
__SEMANTIC_RANGE_INDEX: list[SemanticRangeIndex] = []


def _reset_file_state(uri: str):
    global __SEMANTIC_RANGE_INDEX

    __SEMANTIC_RANGE_INDEX = [
        range_index for range_index in __SEMANTIC_RANGE_INDEX if range_index.uri != uri
    ]

    if uri in __SEMANTIC_TOKEN_INDEX:
        del __SEMANTIC_TOKEN_INDEX[uri]

    if uri in __TEST_RESULTS:
        del __TEST_RESULTS[uri]


def _check_for_duplicate_resources(uri: str):
    counts: defaultdict[str, tuple[int, bool, list[tuple]]] = defaultdict(
        lambda: (0, False, [])
    )

    for resource_uri, resource_key, resource_range, _ in __SEMANTIC_RANGE_INDEX:
        if not constants.RESOURCE_DEF.match(resource_key):
            continue

        count, seen_in_uri, locations = counts[resource_key]
        locations.append((resource_uri, resource_range))
        counts[resource_key] = (
            count + 1,
            seen_in_uri or resource_uri == uri,
            locations,
        )

    duplicate_diagnostics: list[types.Diagnostic] = []

    for resource_key, (count, seen_in_uri, locations) in counts.items():
        if count <= 1 or not seen_in_uri:
            continue

        for resource_uri, resource_range in locations:
            if resource_uri != uri:
                continue

            duplicate_diagnostics.append(
                types.Diagnostic(
                    message=f"Mulitiple instances of {resource_key}.",
                    severity=types.DiagnosticSeverity.Error,
                    range=resource_range,
                )
            )

    return duplicate_diagnostics


def main():
    schema.load_validators_from_files(path=CRD_ROOT)
    server.start_io()


if __name__ == "__main__":
    main()



from typing import NamedTuple

from lsprotocol import types

from koreo import cache
from koreo import registry
from koreo.result import UnwrappedOutcome, is_not_ok, is_unwrapped_ok


from koreo.resource_function.structure import ResourceFunction
from koreo.resource_template.structure import ResourceTemplate
from koreo.value_function.structure import ValueFunction

from koreo.function_test.structure import FunctionTest
from koreo.workflow.structure import Workflow, ErrorStep

from koreo_tooling import constants
from koreo_tooling.function_test import TestResults
from koreo_tooling.indexing.semantics import (
    SemanticAnchor,
    anchor_local_key_search,
    compute_abs_range,
)
from koreo_tooling.langserver.workflow import _process_workflow_step


class HoverResult(NamedTuple):
    hover: types.Hover | None = None
    logs: list[types.LogMessageParams] | None = None


def handle_hover(
    resource_key: str,
    resource_key_range: types.Range,
    local_resource: tuple[str, types.Range] | None,
    test_results: dict[str, TestResults],
):
    match = constants.TOP_LEVEL_RESOURCE.match(resource_key)
    if not match:
        if (
            resource_key.startswith("Workflow:")
            and local_resource
            and local_resource[0].startswith("Step:")
        ):
            return _workflow_step_hover(
                workflow_name=resource_key, local_resource=local_resource
            )

        return HoverResult()

    kind = match.group("kind")
    name = match.group("name")

    if kind == "Workflow":
        return _workflow_hover(
            workflow_name=name,
            resource_key_range=resource_key_range,
        )
    elif kind == "FunctionTest":
        return _function_test_hover(
            test_name=name,
            resource_key_range=resource_key_range,
            test_results=test_results,
        )
    elif kind == "ResourceFunction":
        return _resource_function_hover(
            function_name=name,
            resource_key_range=resource_key_range,
            test_results=test_results,
        )
    elif kind == "ValueFunction":
        return _value_function_hover(
            function_name=name,
            resource_key_range=resource_key_range,
            test_results=test_results,
        )
    elif kind == "ResourceTemplate":
        return _resource_template_hover(
            template_name=name, resource_key_range=resource_key_range
        )

    return HoverResult()


def _workflow_hover(
    workflow_name: str,
    resource_key_range: types.Range,
) -> HoverResult:
    hover_content = [f"# {workflow_name}"]

    workflow = cache.get_resource_from_cache(
        resource_class=Workflow, cache_key=workflow_name
    )

    if not workflow:
        hover_content.append(f"Workflow {workflow_name} not in Koreo Cache")

    elif not is_unwrapped_ok(workflow):
        hover_content.append("Workflow not ready")
        hover_content.append(f"{workflow.message}")

    elif is_not_ok(workflow.steps_ready):
        hover_content.append("Workflow steps not ready")
        hover_content.append(f"{workflow.steps_ready.message}")

    else:
        # TODO: How to validate inputs are OK?
        # This really just means it is in cache, not valid.
        hover_content.append("Workflow prepared successfully.")

    return HoverResult(
        hover=types.Hover(
            contents=types.MarkupContent(
                kind=types.MarkupKind.Markdown,
                value="\n".join(hover_content),
            ),
            range=resource_key_range,
        )
    )


def _workflow_step_hover(
    workflow_name: str,
    local_resource: tuple[str, types.Range],
) -> HoverResult:
    workflow_name = workflow_name.split(":", 1)[-1]

    cached = cache.get_resource_system_data_from_cache(
        resource_class=Workflow, cache_key=workflow_name
    )
    if not cached or not cached.system_data or "anchor" not in cached.system_data:
        return HoverResult()

    step_key, _ = local_resource

    parts = step_key.split(":")
    if not len(parts) == 2:
        return HoverResult()

    _, step_label = parts

    anchor: SemanticAnchor = cached.system_data.get("anchor")
    step_blocks = anchor_local_key_search(
        search_key=step_key, search_nodes=anchor.children
    )

    if len(step_blocks) != 1:
        return HoverResult()

    label_block = anchor_local_key_search(
        search_key=f"label:{step_label}", search_nodes=anchor.children
    )
    if len(label_block) != 1:
        return HoverResult()

    for step in cached.resource.steps:
        if step.label != step_label:
            continue

        if isinstance(step, ErrorStep):
            hover_content = f"Not ready {step.outcome.message}."

        else:
            result = _process_workflow_step(
                step=step,
                step_semantic_block=step_blocks[0],
                step_spec=cached.spec,
                semantic_anchor=anchor,
            )
            message = None
            if result.diagnostics:
                for diagnostic in result.diagnostics:
                    message = diagnostic.message

            if result.error:
                hover_content = "\n".join(["*Error*", message])
            else:
                hover_content = "*Ok*"

        return HoverResult(
            hover=types.Hover(
                contents=types.MarkupContent(
                    kind=types.MarkupKind.Markdown,
                    value=hover_content,
                ),
                range=compute_abs_range(node=label_block[0], anchor=anchor),
            )
        )

    return HoverResult()


def _resource_function_hover(
    function_name: str,
    resource_key_range: types.Range,
    test_results: dict[str, TestResults],
):
    function = cache.get_resource_from_cache(
        resource_class=ResourceFunction, cache_key=function_name
    )

    function_resource = registry.Resource(
        resource_type=ResourceFunction, name=function_name
    )

    return _common_function_hover(
        function_name=function_name,
        function=function,
        registry_resource=function_resource,
        resource_key_range=resource_key_range,
        test_results=test_results,
    )


def _value_function_hover(
    function_name: str,
    resource_key_range: types.Range,
    test_results: dict[str, TestResults],
):
    function = cache.get_resource_from_cache(
        resource_class=ValueFunction, cache_key=function_name
    )

    function_resource = registry.Resource(
        resource_type=ValueFunction, name=function_name
    )

    return _common_function_hover(
        function_name=function_name,
        function=function,
        registry_resource=function_resource,
        resource_key_range=resource_key_range,
        test_results=test_results,
    )


def _common_function_hover(
    function_name: str,
    function: UnwrappedOutcome[ResourceFunction | ValueFunction] | None,
    registry_resource: registry.Resource,
    resource_key_range: types.Range,
    test_results: dict[str, TestResults],
):
    hover_content = [f"# {function_name}"]

    if function is None:
        hover_content.append("*Not cached*")
    elif is_unwrapped_ok(function):
        hover_content.append("Structure is OK")
    else:
        hover_content.append("Function not ready")
        hover_content.append(f"{function.message}")

    tests = {}
    subscribers = registry.get_subscribers(registry_resource)
    for subscriber in subscribers:
        if subscriber.resource_type != FunctionTest.__qualname__:
            continue

        if subscriber.name not in test_results:
            continue

        tests[subscriber.name] = test_results[subscriber.name]

    if tests:
        hover_content.append("## Test Results")
        hover_content.append("| Test | Status | Warnings |")
        hover_content.append("|:-|-:|:-:|")
        for test_name, test_result in tests.items():
            hover_content.append(
                f"| `{test_name}` | {'Pass' if test_result.success else 'Fail'} | {'warnings' if test_result.input_mismatches or test_result.messages else ''} |"
            )

    return HoverResult(
        hover=types.Hover(
            contents=types.MarkupContent(
                kind=types.MarkupKind.Markdown,
                value="\n".join(hover_content),
            ),
            range=resource_key_range,
        )
    )


def _function_test_hover(
    test_name: str,
    resource_key_range: types.Range,
    test_results: dict[str, TestResults],
) -> HoverResult:
    result = test_results.get(test_name)

    if not result:
        return HoverResult(
            hover=types.Hover(
                contents=types.MarkupContent(
                    kind=types.MarkupKind.Markdown, value="*Test has not ran*"
                ),
                range=resource_key_range,
            )
        )

    hover_content = []

    if result.success:
        hover_content.append(f"# *Success*")

        if result.messages:
            hover_content.append("## Test Notes")
            hover_content.extend(result.messages)

        if result.input_mismatches:
            hover_content.append("## Unused Inputs")
            hover_content.append("| Field | Issue | Severity |")
            hover_content.append("|:-|-:|:-:|")
            for mismatch in result.input_mismatches:
                hover_content.append(
                    f"| `{mismatch.field}` | {_input_error_formatter(actual=mismatch.actual, expected=mismatch.expected)} | {mismatch.severity} |"
                )
            hover_content.append("\n")

    else:
        hover_content.append("# *Error*")
        if result.messages:
            hover_content.append("## Failure")
            hover_content.extend(result.messages)

        if result.input_mismatches:
            hover_content.append("## Input Mismatches")
            hover_content.append("| Field | Issue | Severity |")
            hover_content.append("|:-|-:|:-:|")
            for mismatch in result.input_mismatches:
                hover_content.append(
                    f"| `{mismatch.field}` | {_input_error_formatter(actual=mismatch.actual, expected=mismatch.expected)} | {mismatch.severity} |"
                )
            hover_content.append("\n")

        if result.resource_field_errors:
            hover_content.append("## Field Mismatches")
            hover_content.append("| Field | Actual | Expected |")
            hover_content.append("|:-|-:|-:|")
            for compare in result.resource_field_errors:
                hover_content.append(
                    f"| `{compare.field}` | {compare.actual} | {compare.expected} |"
                )
            hover_content.append("\n")

        if result.outcome_fields_errors:
            hover_content.append("## Outcome Mismatches")
            hover_content.append("| Field | Actual | Expected |")
            hover_content.append("|:-|-:|-:|")
            for compare in result.outcome_fields_errors:
                hover_content.append(
                    f"| `{compare.field}` | {compare.actual} | {compare.expected} |"
                )
            hover_content.append("\n")

    return HoverResult(
        hover=types.Hover(
            contents=types.MarkupContent(
                kind=types.MarkupKind.Markdown,
                value="\n".join(hover_content),
            ),
            range=resource_key_range,
        )
    )


def _resource_template_hover(
    template_name: str,
    resource_key_range: types.Range,
) -> HoverResult:
    hover_content = [f"# {template_name}"]

    resource_template = cache.get_resource_from_cache(
        resource_class=ResourceTemplate, cache_key=template_name
    )

    if not resource_template:
        hover_content.append(f"ResourceTemplate {template_name} not in Koreo Cache")

    elif not is_unwrapped_ok(resource_template):
        hover_content.append("Resource template not ready")
        hover_content.append(f"{resource_template.message}")

    else:
        hover_content.append("ResourceTemplate prepared successfully.")

    return HoverResult(
        hover=types.Hover(
            contents=types.MarkupContent(
                kind=types.MarkupKind.Markdown,
                value="\n".join(hover_content),
            ),
            range=resource_key_range,
        )
    )


def _input_error_formatter(actual: bool, expected: bool) -> str:
    if actual and not expected:
        return "unexpected"

    if not actual and expected:
        return "*missing*"

    return "_unknown_"



from typing import Sequence

from lsprotocol import types

from koreo_tooling.indexing.semantics import (
    SemanticAnchor,
    SemanticBlock,
    SemanticNode,
    anchor_local_key_search,
    compute_abs_range,
)


type SemanticReturn = SemanticAnchor | SemanticBlock | SemanticNode
type RangerError = list[types.Diagnostic]
type RangerReturns = SemanticReturn | RangerError | None


def block_range_extract(
    search_key: str,
    search_nodes: Sequence[SemanticAnchor | SemanticBlock | SemanticNode] | None,
    anchor: SemanticAnchor,
) -> RangerReturns:
    matches = anchor_local_key_search(search_key, search_nodes=search_nodes)

    if not matches:
        return None

    match, *extras = matches

    if extras:
        return [
            types.Diagnostic(
                message=f"Multiple instances of ('{search_key}').",
                severity=types.DiagnosticSeverity.Error,
                range=compute_abs_range(match, anchor),
            )
            for match in matches
        ]

    return match


def key_value_range_extract(
    search_key: str,
    search_nodes: Sequence[SemanticAnchor | SemanticBlock | SemanticNode] | None,
    anchor: SemanticAnchor,
) -> RangerReturns:
    key = block_range_extract(
        search_key=search_key, search_nodes=search_nodes, anchor=anchor
    )
    match key:
        case None:
            return None
        case list(block_diagnostics):
            return block_diagnostics

    if not key.children:
        return [
            types.Diagnostic(
                message=f"Missing value for ('{search_key}').",
                severity=types.DiagnosticSeverity.Warning,
                range=compute_abs_range(key, anchor),
            )
        ]

    value, *extras = key.children

    if extras:
        return [
            types.Diagnostic(
                message=f"Multiple children of ('{search_key}'), child {idx}.",
                severity=types.DiagnosticSeverity.Error,
                range=compute_abs_range(child, anchor),
            )
            for idx, child in enumerate(key.children)
        ]

    return value


def nested_range_extract(
    search_keys: list[str], search_nodes: Sequence, anchor: SemanticAnchor
) -> RangerReturns:
    while search_keys:
        search_key, *search_keys = search_keys
        next_node = block_range_extract(
            search_key=search_key,
            search_nodes=search_nodes,
            anchor=anchor,
        )
        match next_node:
            case None:
                return None
            case list(block_diagnostics):
                return block_diagnostics

        if not search_keys:
            return next_node

        if not next_node.children:
            return None

        search_nodes = next_node.children

    return None



from typing import Any, Awaitable, Callable, NamedTuple, Protocol
import asyncio


class KillRequest:
    pass


class ProccessRequest(NamedTuple):
    monotime: float
    response: asyncio.Queue[Exception | Any]


class FileProcessor(Protocol):
    async def __call__(self, file_uri: str) -> Awaitable: ...


class HandlerFailure(Exception):
    pass


async def handle_file(file_uri: str, monotime: float, file_processor: FileProcessor):
    worker_queue = _setup_worker(file_uri=file_uri, file_processor=file_processor)

    response_queue = asyncio.Queue(10)

    await worker_queue.put(ProccessRequest(monotime=monotime, response=response_queue))

    result = await response_queue.get()
    response_queue.task_done()

    if isinstance(result, Exception):
        raise result

    return result


async def shutdown_handlers():
    async with asyncio.TaskGroup() as task_group:
        for queue in _FILE_HANDLER_QUEUES.values():
            task_group.create_task(queue.put(KillRequest()))


def _setup_worker(
    file_uri: str, file_processor: FileProcessor
) -> asyncio.LifoQueue[ProccessRequest | KillRequest]:
    handler_key = f"{file_processor.__qualname__}${file_uri}"

    if handler_key in _FILE_HANDLER_QUEUES:
        return _FILE_HANDLER_QUEUES[handler_key]

    file_handler_queue = asyncio.LifoQueue[ProccessRequest | KillRequest]()
    _FILE_HANDLER_QUEUES[handler_key] = file_handler_queue

    handler_task = asyncio.create_task(
        _file_processor_handler(
            handler_key=handler_key, file_uri=file_uri, file_processor=file_processor
        ),
        name=handler_key,
    )
    _FILE_HANDLERS[handler_key] = handler_task

    handler_task.add_done_callback(_cleanup_file_handler)

    return file_handler_queue


_FILE_HANDLER_QUEUES: dict[str, asyncio.LifoQueue[ProccessRequest | KillRequest]] = {}
_FILE_HANDLERS: dict[str, asyncio.Task] = {}


def _cleanup_file_handler(file_handler: asyncio.Task):
    handler_key = file_handler.get_name()

    if handler_key in _FILE_HANDLERS:
        del _FILE_HANDLERS[handler_key]

    if handler_key in _FILE_HANDLER_QUEUES:
        queue = _FILE_HANDLER_QUEUES[handler_key]
        del _FILE_HANDLER_QUEUES[handler_key]

        err = HandlerFailure(f"Exception within handler for '{handler_key}'")

        # Unblock any processes waiting for this defunct handler.
        while not queue.empty():
            request = queue.get_nowait()
            try:
                match request:
                    case KillRequest():
                        continue
                    case ProccessRequest(response=response_queue):
                        response_queue.put_nowait(err)
            finally:
                queue.task_done()


async def _file_processor_handler(
    handler_key: str, file_uri: str, file_processor: Callable[[str, float], Awaitable]
):
    if handler_key not in _FILE_HANDLER_QUEUES:
        raise RuntimeError(
            f"{file_processor.__qualname__} for ('{file_uri}') not initialized."
        )

    handler_queue = _FILE_HANDLER_QUEUES[handler_key]

    last_monotime = 0
    last_result = None
    while True:
        work_request = await handler_queue.get()
        try:
            match work_request:
                case KillRequest():
                    break
                case ProccessRequest(
                    monotime=monotime, response=response_queue
                ) if monotime <= last_monotime:
                    await response_queue.put(last_result)
                    continue

                case ProccessRequest(monotime=monotime, response=response_queue):
                    try:
                        result = await file_processor(file_uri)
                        await response_queue.put(result)

                        last_monotime = monotime
                        last_result = result
                    except Exception as err:
                        await response_queue.put(err)
        finally:
            handler_queue.task_done()



import unittest

from koreo_tooling.indexing.cel_semantics import (
    NodeDiagnostic,
    SemanticNode,
    Position,
    Severity,
    parse,
)


class TestParse(unittest.TestCase):
    def test_empty_cel(self):
        anchor_base_pos = Position(line=0, character=0)
        nodes = parse([""], anchor_base_pos=anchor_base_pos)
        self.assertListEqual([], nodes)

    def test_simple_number(self):
        anchor_base_pos = Position(line=0, character=0)
        nodes = parse(["1"], anchor_base_pos=anchor_base_pos)

        expected = [
            SemanticNode(
                position=Position(line=0, character=0),
                anchor_rel=anchor_base_pos,
                length=1,
                node_type="number",
                modifier=[],
            )
        ]

        self.assertListEqual(expected, nodes)

    def test_operator(self):
        anchor_base_pos = Position(line=0, character=0)
        nodes = parse(["+"], anchor_base_pos=anchor_base_pos)

        expected = [
            SemanticNode(
                position=Position(line=0, character=0),
                anchor_rel=Position(line=0, character=0),
                length=1,
                node_type="operator",
                modifier=[],
            ),
        ]

        self.assertListEqual(expected, nodes)

    def test_symbol(self):
        anchor_base_pos = Position(line=0, character=0)
        nodes = parse(["inputs"], anchor_base_pos=anchor_base_pos)

        expected = [
            SemanticNode(
                position=Position(line=0, character=0),
                anchor_rel=Position(line=0, character=0),
                length=6,
                node_type="variable",
                modifier=[],
            ),
        ]

        self.assertListEqual(expected, nodes)

    def test_quoted(self):
        anchor_base_pos = Position(line=0, character=0)
        nodes = parse(["'this is a lot'"], anchor_base_pos=anchor_base_pos)

        expected = [
            SemanticNode(
                position=Position(line=0, character=0),
                anchor_rel=Position(line=0, character=0),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=0, character=1),
                length=13,
                node_type="string",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=13),
                anchor_rel=Position(line=0, character=14),
                length=1,
                node_type="operator",
                modifier=[],
            ),
        ]

        self.assertListEqual(expected, nodes)

    def test_simple_formula(self):
        anchor_base_pos = Position(line=0, character=0)
        nodes = parse(["1 + 1"], anchor_base_pos=anchor_base_pos)

        expected = [
            SemanticNode(
                position=Position(line=0, character=0),
                anchor_rel=Position(line=0, character=0),
                length=1,
                node_type="number",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=0, character=2),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=0, character=4),
                length=1,
                node_type="number",
                modifier=[],
            ),
        ]

        self.assertListEqual(expected, nodes)

    def test_mismatched_quote(self):
        anchor_base_pos = Position(line=0, character=0)
        with self.assertRaises(RuntimeError):
            parse(["'"], anchor_base_pos=anchor_base_pos)

        with self.assertRaises(RuntimeError):
            parse(['"'], anchor_base_pos=anchor_base_pos)

    def test_seed_offset_multiline(self):
        anchor_base_pos = Position(line=10, character=0)
        nodes = parse(
            ["1", "      +", "      1", ""],
            seed_offset=15,
            abs_offset=5,
            anchor_base_pos=anchor_base_pos,
        )

        expected = [
            SemanticNode(
                position=Position(line=0, character=15),
                anchor_rel=Position(line=10, character=20),
                length=1,
                node_type="number",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=1, character=6),
                anchor_rel=Position(line=11, character=6),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=1, character=6),
                anchor_rel=Position(line=12, character=6),
                length=1,
                node_type="number",
                modifier=[],
            ),
        ]

        self.maxDiff = None
        self.assertListEqual(expected, nodes)

    def test_seed_line_multiline(self):
        anchor_base_pos = Position(line=5, character=0)
        nodes = parse(
            ["      1", "      + ", "      1 ", ""],
            seed_line=2,
            anchor_base_pos=anchor_base_pos,
        )

        expected = [
            SemanticNode(
                position=Position(line=2, character=6),
                anchor_rel=Position(line=7, character=6),
                length=1,
                node_type="number",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=1, character=6),
                anchor_rel=Position(line=8, character=6),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=1, character=6),
                anchor_rel=Position(line=9, character=6),
                length=1,
                node_type="number",
                modifier=[],
            ),
        ]

        self.maxDiff = None
        self.assertListEqual(expected, nodes)

    def test_multiline_with_extra_newlines(self):
        anchor_base_pos = Position(line=13, character=0)
        nodes = parse(
            [
                "      1",
                "",
                "",
                "",
                "      +",
                "",
                "",
                "       ",
                "",
                "",
                "      1",
                "",
            ],
            seed_line=1,
            anchor_base_pos=anchor_base_pos,
        )

        expected = [
            SemanticNode(
                position=Position(line=1, character=6),
                anchor_rel=Position(line=14, character=6),
                length=1,
                node_type="number",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=4, character=6),
                anchor_rel=Position(line=18, character=6),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=6, character=6),
                anchor_rel=Position(line=24, character=6),
                length=1,
                node_type="number",
                modifier=[],
            ),
        ]

        self.maxDiff = None
        self.assertListEqual(expected, nodes)

    def test_trailing_comma_single_line(self):
        anchor_base_pos = Position(line=1, character=0)
        nodes = parse(['{"key": value,  }'], anchor_base_pos=anchor_base_pos)
        expected = [
            # {
            SemanticNode(
                position=Position(line=0, character=0),
                anchor_rel=Position(line=1, character=0),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=1, character=1),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=1, character=2),
                length=3,
                node_type="property",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=1, character=5),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=1, character=6),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=1, character=8),
                length=5,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=5),
                anchor_rel=Position(line=1, character=13),
                length=1,
                node_type="operator",
                modifier=[],
                diagnostic=NodeDiagnostic(
                    message="Trailing commas are unsupported.", severity=Severity.error
                ),
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=1, character=16),
                length=1,
                node_type="operator",
                modifier=[],
            ),
        ]

        self.maxDiff = None
        self.assertListEqual(expected, nodes)

    def test_trailing_comma_multi_line(self):
        anchor_base_pos = Position(line=2, character=0)
        nodes = parse(
            ["{", '  "key": value,', "}", ""], anchor_base_pos=anchor_base_pos
        )
        expected = [
            # {
            SemanticNode(
                position=Position(line=0, character=0),
                anchor_rel=Position(line=2, character=0),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=1, character=2),
                anchor_rel=Position(line=3, character=2),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=3, character=3),
                length=3,
                node_type="property",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=3, character=6),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=3, character=7),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=3, character=9),
                length=5,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=5),
                anchor_rel=Position(line=3, character=14),
                length=1,
                node_type="operator",
                modifier=[],
                diagnostic=NodeDiagnostic(
                    message="Trailing commas are unsupported.", severity=Severity.error
                ),
            ),
            SemanticNode(
                position=Position(line=1, character=0),
                anchor_rel=Position(line=4, character=0),
                length=1,
                node_type="operator",
                modifier=[],
            ),
        ]

        self.maxDiff = None
        self.assertListEqual(expected, nodes)

    def test_complex_white_space(self):
        anchor_base_pos = Position(line=0, character=0)
        nodes = parse(
            ["    int('1717' )            +    9"], anchor_base_pos=anchor_base_pos
        )

        expected = [
            SemanticNode(
                position=Position(line=0, character=4),
                anchor_rel=Position(line=0, character=4),
                length=3,
                node_type="function",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=0, character=7),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=0, character=8),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=0, character=9),
                length=4,
                node_type="string",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=4),
                anchor_rel=Position(line=0, character=13),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=0, character=15),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=13),
                anchor_rel=Position(line=0, character=28),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=5),
                anchor_rel=Position(line=0, character=33),
                length=1,
                node_type="number",
                modifier=[],
            ),
        ]

        self.maxDiff = None
        self.assertListEqual(expected, nodes)

    def test_complex_multiline(self):
        anchor_base_pos = Position(line=3, character=0)
        tokens = parse(
            [
                "",
                "{",
                "  \"complicated.key.name\": 'value',",
                '  unquoted: "key",',
                '  "formula": 1 + 812,',
                "  function: a.name(),",
                '  "index": avar[2] + avar["key"],',
                '  "entry": inputs.map(key, {key: 22})',
                "}",
                "",
            ],
            anchor_base_pos=anchor_base_pos,
        )

        expected = [
            # {
            SemanticNode(
                position=Position(line=1, character=0),
                anchor_rel=Position(line=4, character=0),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            #   "complicated.key.name": 'value',
            SemanticNode(
                position=Position(line=1, character=2),
                anchor_rel=Position(line=5, character=2),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=5, character=3),
                length=20,
                node_type="property",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=20),
                anchor_rel=Position(line=5, character=23),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=5, character=24),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=5, character=26),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=5, character=27),
                length=5,
                node_type="string",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=5),
                anchor_rel=Position(line=5, character=32),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=5, character=33),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            #   unquoted: "key",
            SemanticNode(
                position=Position(line=1, character=2),
                anchor_rel=Position(line=6, character=2),
                length=8,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=8),
                anchor_rel=Position(line=6, character=10),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=6, character=12),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=6, character=13),
                length=3,
                node_type="string",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=6, character=16),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=6, character=17),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            #   "formula": 1 + 8,
            SemanticNode(
                position=Position(line=1, character=2),
                anchor_rel=Position(line=7, character=2),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=7, character=3),
                length=7,
                node_type="property",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=7),
                anchor_rel=Position(line=7, character=10),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=7, character=11),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=7, character=13),
                length=1,
                node_type="number",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=7, character=15),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=7, character=17),
                length=3,
                node_type="number",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=7, character=20),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            #   function: a.name()
            SemanticNode(
                position=Position(line=1, character=2),
                anchor_rel=Position(line=8, character=2),
                length=8,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=8),
                anchor_rel=Position(line=8, character=10),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=8, character=12),
                length=1,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=8, character=13),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=8, character=14),
                length=4,
                node_type="function",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=4),
                anchor_rel=Position(line=8, character=18),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=8, character=19),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=8, character=20),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            #   "index": avar[2] + avar["key"]
            SemanticNode(
                position=Position(line=1, character=2),
                anchor_rel=Position(line=9, character=2),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=9, character=3),
                length=5,
                node_type="property",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=5),
                anchor_rel=Position(line=9, character=8),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=9, character=9),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=9, character=11),
                length=4,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=4),
                anchor_rel=Position(line=9, character=15),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=9, character=16),
                length=1,
                node_type="number",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=9, character=17),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=9, character=19),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=9, character=21),
                length=4,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=4),
                anchor_rel=Position(line=9, character=25),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=9, character=26),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=9, character=27),
                length=3,
                node_type="string",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=9, character=30),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=9, character=31),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=9, character=32),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            #   "entry": inputs.map(key, {key: 22})
            SemanticNode(
                position=Position(line=1, character=2),
                anchor_rel=Position(line=10, character=2),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=10, character=3),
                length=5,
                node_type="property",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=5),
                anchor_rel=Position(line=10, character=8),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=10, character=9),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=10, character=11),
                length=6,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=6),
                anchor_rel=Position(line=10, character=17),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=10, character=18),
                length=3,
                node_type="keyword",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=10, character=21),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=10, character=22),
                length=3,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=10, character=25),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=10, character=27),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=10, character=28),
                length=3,
                node_type="variable",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=3),
                anchor_rel=Position(line=10, character=31),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=10, character=33),
                length=2,
                node_type="number",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=2),
                anchor_rel=Position(line=10, character=35),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            SemanticNode(
                position=Position(line=0, character=1),
                anchor_rel=Position(line=10, character=36),
                length=1,
                node_type="operator",
                modifier=[],
            ),
            # }
            SemanticNode(
                position=Position(line=1, character=0),
                anchor_rel=Position(line=11, character=0),
                length=1,
                node_type="operator",
                modifier=[],
            ),
        ]

        self.maxDiff = None
        self.assertListEqual(expected, tokens)



from __future__ import annotations
from dataclasses import dataclass
from typing import (
    Any,
    Literal,
    NamedTuple,
    Protocol,
    Sequence,
    get_args,
)
import enum

from lsprotocol.types import Position, Range

TokenType = Literal[
    "",
    "argument",
    "class",
    "comment",
    "decorator",
    "enum",
    "enumMember",
    "event",
    "function",
    "interface",
    "keyword",
    "macro",
    "method",
    "modifier",
    "namespace",
    "number",
    "operator",
    "parameter",
    "property",
    "regexp",
    "string",
    "struct",
    "type",
    "typeParameter",
    "variable",
]

TokenTypes = get_args(TokenType)


class Modifier(enum.IntFlag):
    declaration = enum.auto()
    definition = enum.auto()
    readonly = enum.auto()
    static = enum.auto()
    deprecated = enum.auto()
    abstract = enum.auto()
    modification = enum.auto()
    documentation = enum.auto()
    defaultLibrary = enum.auto()


TokenModifiers: Sequence[str] = [modifier.name for modifier in Modifier]


class Severity(enum.IntFlag):
    debug = enum.auto()
    info = enum.auto()
    warning = enum.auto()
    error = enum.auto()


class NodeDiagnostic(NamedTuple):
    message: str
    severity: Severity


class SemanticAnchor(NamedTuple):
    key: str | None
    abs_position: Position
    rel_position: Position
    children: Sequence[SemanticBlock | SemanticNode] | None = None


class SemanticBlock(NamedTuple):
    local_key: str | None
    index_key: str | None
    anchor_rel: Range
    children: Sequence[SemanticBlock | SemanticNode] | None = None


class SemanticNode(NamedTuple):
    position: Position
    anchor_rel: Position
    length: int
    local_key: str | None = None
    index_key: str | None = None
    node_type: TokenType = ""
    modifier: list[Modifier] | None = None
    children: Sequence[SemanticBlock | SemanticNode] | None = None
    diagnostic: NodeDiagnostic | None = None


class IndexFn(Protocol):
    def __call__(self, value: Any) -> str | None: ...


class FieldIndexFn(Protocol):
    def __call__(self, value: Any) -> tuple[str, str] | None: ...


type SemanticStructureMap = dict[str, SemanticStructure]


@dataclass
class SemanticStructure:
    type: TokenType = ""
    modifier: list[Modifier] | None = None
    local_key_fn: IndexFn | None = None
    index_key_fn: IndexFn | None = None
    field_index_key_fn: FieldIndexFn | None = None
    sub_structure: SemanticStructureMap | SemanticStructure | None = None
    strict_sub_structure_keys: bool = False


def flatten(
    nodes: (
        SemanticAnchor
        | SemanticBlock
        | SemanticNode
        | Sequence[SemanticAnchor | SemanticNode]
    ),
) -> Sequence[SemanticNode]:
    if isinstance(nodes, (SemanticAnchor, SemanticBlock, SemanticNode)):
        return flatten_node(nodes)

    flattened = []

    for node in nodes:
        flattened.extend(flatten_node(node))

    return flattened


def flatten_node(
    node: SemanticAnchor | SemanticBlock | SemanticNode,
) -> Sequence[SemanticNode]:
    flattened = []
    if isinstance(node, SemanticNode):
        flattened.append(node._replace(children=None))

    if not node.children:
        return flattened

    for child_node in node.children:
        flattened.extend(flatten_node(child_node))

    return flattened


def extract_diagnostics(nodes: Sequence[SemanticNode]) -> Sequence[SemanticNode]:
    return [node for node in nodes if node.diagnostic]


def generate_key_range_index(
    nodes: (
        SemanticAnchor
        | SemanticBlock
        | SemanticNode
        | Sequence[SemanticBlock | SemanticNode]
    ),
    anchor: SemanticAnchor | None = None,
) -> Sequence[tuple[str, Range]]:
    index = []

    match nodes:
        case SemanticAnchor(key=key, children=children):
            if key:
                index.append((key, compute_abs_range(nodes, anchor=nodes)))

            if children:
                index.extend(generate_key_range_index(nodes=children, anchor=nodes))

            return index

        case SemanticBlock(index_key=index_key, children=children):
            if index_key and anchor:
                index.append((index_key, compute_abs_range(nodes, anchor=anchor)))

            if children:
                index.extend(generate_key_range_index(nodes=children, anchor=anchor))

            return index

        case SemanticNode(index_key=index_key, children=children):
            if index_key and anchor:
                index.append((index_key, compute_abs_range(nodes, anchor=anchor)))

            if children:
                index.extend(generate_key_range_index(nodes=children, anchor=anchor))

            return index

    for node in nodes:
        index.extend(generate_key_range_index(nodes=node, anchor=anchor))

    return index


def generate_local_range_index(
    nodes: (
        SemanticAnchor
        | SemanticBlock
        | SemanticNode
        | Sequence[SemanticBlock | SemanticNode]
        | None
    ),
    anchor: SemanticAnchor | None = None,
) -> Sequence[tuple[str, Range]]:
    if not nodes:
        return []

    index = []

    match nodes:
        case SemanticAnchor(key=key, children=children):
            if key:
                index.append((key, compute_abs_range(nodes, anchor=nodes)))

            if children:
                index.extend(generate_local_range_index(nodes=children, anchor=nodes))

            return index

        case SemanticBlock(local_key=local_key, children=children):
            if local_key and anchor:
                index.append((local_key, compute_abs_range(nodes, anchor=anchor)))

            if children:
                index.extend(generate_local_range_index(nodes=children, anchor=anchor))

            return index

        case SemanticNode(local_key=local_key, children=children):
            if local_key and anchor:
                index.append((local_key, compute_abs_range(nodes, anchor=anchor)))

            if children:
                index.extend(generate_local_range_index(nodes=children, anchor=anchor))

            return index

    for node in nodes:
        index.extend(generate_local_range_index(nodes=node, anchor=anchor))

    return index


def anchor_local_key_search(
    search_key: str,
    search_nodes: Sequence[SemanticAnchor | SemanticBlock | SemanticNode] | None = None,
) -> Sequence[SemanticAnchor | SemanticBlock | SemanticNode]:
    if not search_nodes or not search_key:
        return []

    index = []
    for node in search_nodes:
        match node:
            case SemanticAnchor(key=key):
                if key and key == search_key:
                    index.append(node)

            case SemanticBlock(local_key=local_key):
                if local_key and local_key == search_key:
                    index.append(node)

            case SemanticNode(local_key=local_key):
                if search_key and search_key == local_key:
                    index.append(node)

        if not node.children:
            continue

        index.extend(
            anchor_local_key_search(
                search_key=search_key,
                search_nodes=node.children,
            )
        )

    return index


def compute_abs_position(
    rel_position: Position, abs_position: Position, length: int = 0
) -> Position:
    return Position(
        line=abs_position.line + rel_position.line,
        character=rel_position.character + length,
    )


def compute_abs_range(
    node: SemanticAnchor | SemanticBlock | SemanticNode, anchor: SemanticAnchor
) -> Range:
    match node:
        case SemanticAnchor(abs_position=abs_position):
            end_node = node
            while True:
                if not end_node.children:
                    break
                end_node = end_node.children[-1]

            match end_node:
                case SemanticAnchor(abs_position=abs_position):
                    end_pos = Position(
                        line=abs_position.line,
                        character=abs_position.character,
                    )
                case SemanticBlock(anchor_rel=anchor_rel):
                    end_pos = compute_abs_position(
                        anchor_rel.end, abs_position=abs_position
                    )
                case SemanticNode(anchor_rel=anchor_rel, length=length):
                    end_pos = compute_abs_position(
                        anchor_rel, abs_position=abs_position, length=length
                    )

            return Range(
                start=Position(
                    line=abs_position.line,
                    character=abs_position.character,
                ),
                end=end_pos,
            )

        case SemanticNode(anchor_rel=anchor_rel, length=length):
            return Range(
                start=Position(
                    line=anchor.abs_position.line + anchor_rel.line,
                    character=anchor_rel.character,
                ),
                end=Position(
                    line=anchor.abs_position.line + anchor_rel.line,
                    character=anchor_rel.character + length,
                ),
            )

        case SemanticBlock(anchor_rel=anchor_rel):
            return Range(
                start=Position(
                    line=anchor.abs_position.line + anchor_rel.start.line,
                    character=anchor_rel.start.character,
                ),
                end=Position(
                    line=anchor.abs_position.line + anchor_rel.end.line,
                    character=anchor_rel.end.character,
                ),
            )

    # This should be impossible
    raise Exception(f"Bad node type ({type(node)})")



from typing import Iterable


def call_arg_compare(
    lhs: Iterable[str], rhs: Iterable[str]
) -> dict[str, tuple[bool, bool]]:
    arguments: dict[str, tuple[bool, bool]] = {}
    for argument in lhs:
        arguments[argument] = (True, False)

    for argument in rhs:
        lhs_state, _ = arguments.get(argument, (False, False))
        arguments[argument] = (lhs_state, True)

    return arguments



from typing import NamedTuple, Sequence
import os

os.environ["KOREO_DEV_TOOLING"] = "true"


from lsprotocol import types

from koreo import cache
from koreo.workflow.structure import ErrorStep, LogicSwitch, Step, Workflow
from koreo.resource_function.structure import ResourceFunction
from koreo.value_function.structure import ValueFunction
from koreo.result import is_unwrapped_ok

from koreo_tooling import constants
from koreo_tooling.analysis import call_arg_compare
from koreo_tooling.indexing import compute_abs_range
from koreo_tooling.langserver.rangers import block_range_extract


def process_workflows(
    uri: str, workflows: Sequence[tuple[str, types.Range]]
) -> list[types.Diagnostic]:
    diagnostics: list[types.Diagnostic] = []

    for koreo_cache_key, resource_range in workflows:
        cached_resource = cache.get_resource_system_data_from_cache(
            resource_class=Workflow, cache_key=koreo_cache_key
        )

        if not cached_resource or not cached_resource.resource:
            diagnostics.append(
                types.Diagnostic(
                    message=f"Workflow not yet prepared ({koreo_cache_key}) or not in Koreo cache.",
                    severity=types.DiagnosticSeverity.Error,
                    range=resource_range,
                )
            )
            continue

        if not is_unwrapped_ok(cached_resource.resource):
            diagnostics.append(
                types.Diagnostic(
                    message=f"Workflow ({koreo_cache_key}) is not ready ({cached_resource.resource.message}).",
                    severity=types.DiagnosticSeverity.Error,
                    range=resource_range,
                )
            )
            continue

        workflow_result = _process_workflow(
            uri=uri,
            resource_range=resource_range,
            workflow_name=koreo_cache_key,
            workflow=cached_resource.resource,
            raw_spec=cached_resource.spec,
            koreo_metadata=cached_resource.system_data,
        )
        if workflow_result.diagnostics:
            diagnostics.extend(workflow_result.diagnostics)

    return diagnostics


class ProcessResult(NamedTuple):
    error: bool = False
    diagnostics: list[types.Diagnostic] | None = None


def _process_workflow(
    uri: str,
    resource_range: types.Range,
    workflow_name: str,
    workflow: Workflow,
    raw_spec: dict,
    koreo_metadata: dict | None,
) -> ProcessResult:
    if not koreo_metadata:
        return ProcessResult(
            error=True,
            diagnostics=[
                types.Diagnostic(
                    message=f"Workflow ('{workflow_name}') missing in Koreo Cache (this _should_ be impossible).",
                    severity=types.DiagnosticSeverity.Error,
                    range=resource_range,
                )
            ],
        )

    semantic_uri = koreo_metadata.get("uri", "")
    if semantic_uri != uri:
        return ProcessResult(
            error=True,
            diagnostics=[
                types.Diagnostic(
                    message=(
                        f"Duplicate Workflow ('{workflow_name}') detected in "
                        f"('{semantic_uri}'), skipping further analysis."
                    ),
                    severity=types.DiagnosticSeverity.Error,
                    range=resource_range,
                )
            ],
        )

    semantic_anchor = koreo_metadata.get("anchor")
    if not semantic_anchor:
        return ProcessResult(
            error=True,
            diagnostics=[
                types.Diagnostic(
                    message=f"Unknown error processing Workflow ('{workflow_name}'), semantic analysis data missing from Koreo cache.",
                    severity=types.DiagnosticSeverity.Error,
                    range=resource_range,
                )
            ],
        )

    diagnostics: list[types.Diagnostic] = []
    if not is_unwrapped_ok(workflow.steps_ready):
        diagnostics.append(
            types.Diagnostic(
                message=f"Workflow is not ready ({workflow.steps_ready.message}).",
                severity=types.DiagnosticSeverity.Warning,
                range=resource_range,
            )
        )
        if not raw_spec:
            return ProcessResult(error=True, diagnostics=diagnostics)

    raw_steps_spec = raw_spec.get("steps")

    if not raw_steps_spec:
        step_specs = {}
    else:
        step_specs = {step_spec.get("label"): step_spec for step_spec in raw_steps_spec}

    if not step_specs and workflow.steps:
        diagnostics.append(
            types.Diagnostic(
                message=f"Workflow steps are malformed.",
                severity=types.DiagnosticSeverity.Warning,
                range=resource_range,
            )
        )
        return ProcessResult(error=True, diagnostics=diagnostics)

    has_step_error = False

    for step in workflow.steps:
        step_block = block_range_extract(
            search_key=f"Step:{step.label}",
            search_nodes=semantic_anchor.children,
            anchor=semantic_anchor,
        )
        match step_block:
            case None:
                continue
            case list(block_diagnostics):
                diagnostics.extend(block_diagnostics)
                continue

        step_spec = step_specs.get(step.label)

        step_result = _process_workflow_step(
            step, step_block, step_spec, semantic_anchor
        )

        has_step_error = has_step_error or step_result.error
        if step_result.diagnostics:
            diagnostics.extend(step_result.diagnostics)

    if has_step_error:
        diagnostics.append(
            types.Diagnostic(
                message=f"Workflow steps are not ready.",
                severity=types.DiagnosticSeverity.Error,
                range=resource_range,
            )
        )

    return ProcessResult(error=has_step_error, diagnostics=diagnostics)


def _process_workflow_step(
    step: Step | ErrorStep,
    step_semantic_block,
    step_spec,
    semantic_anchor,
    implicit_inputs: Sequence[str] | None = None,
) -> ProcessResult:
    if isinstance(step, ErrorStep):
        return ProcessResult(
            error=True,
            diagnostics=_step_label_error_diagnostic(
                label=step.label,
                step_semantic_block=step_semantic_block,
                semantic_anchor=semantic_anchor,
                message=f"Not ready {step.outcome.message}.",
            ),
        )

    first_tier_inputs = _get_first_tier_inputs(step.logic)

    raw_inputs = step_spec.get("inputs", {})

    provided_input_keys = list(raw_inputs.keys())

    if implicit_inputs:
        provided_input_keys.extend(implicit_inputs)

    raw_for_each = step_spec.get("forEach", {})
    for_each_key = raw_for_each.get("inputKey")
    if for_each_key:
        provided_input_keys.append(f"{for_each_key}")

    has_error = False
    diagnostics: list[types.Diagnostic] = []

    inputs_block = block_range_extract(
        search_key="inputs",
        search_nodes=step_semantic_block.children,
        anchor=semantic_anchor,
    )
    match inputs_block:
        case list(block_diagnostics):
            has_error = True
            diagnostics.extend(block_diagnostics)
            inputs_block = None

    inputs = call_arg_compare(provided_input_keys, first_tier_inputs)
    for argument, (provided, expected) in inputs.items():
        if not expected and provided:
            has_error = True

            # This is not done earlier so that we can still flag issues at the
            # workflow level.
            if not inputs_block:
                continue

            input_block = block_range_extract(
                search_key=f"input:{argument}",
                search_nodes=step_semantic_block.children,
                anchor=semantic_anchor,
            )
            match input_block:
                case None:
                    continue
                case list(block_diagnostics):
                    diagnostics.extend(block_diagnostics)
                    continue

            raw_arg = raw_inputs.get(argument)

            diagnostics.append(
                types.Diagnostic(
                    message=f"Input ('{argument}') not expected. {raw_arg}.",
                    severity=types.DiagnosticSeverity.Warning,
                    range=compute_abs_range(input_block, semantic_anchor),
                )
            )

        elif expected and not provided:
            has_error = True

            # This is not done earlier so that we can still flag issues at the
            # workflow level.
            if not inputs_block:
                continue

            diagnostics.append(
                types.Diagnostic(
                    message=f"Missing input ('{argument}').",
                    severity=types.DiagnosticSeverity.Error,
                    range=compute_abs_range(inputs_block, semantic_anchor),
                )
            )

    if has_error:
        diagnostics.extend(
            _step_label_error_diagnostic(
                label=step.label,
                step_semantic_block=step_semantic_block,
                semantic_anchor=semantic_anchor,
                message="Step has errors",
            )
        )
    return ProcessResult(error=has_error, diagnostics=diagnostics)


def _get_first_tier_inputs(
    logic: Workflow | ValueFunction | ResourceFunction | LogicSwitch,
) -> set[str]:
    match logic:
        case (LogicSwitch() | ValueFunction() | ResourceFunction()) as logic:
            # These are just the "top level" direct inputs. No consideration to
            # internal structure.
            return set(
                input_key.group("name")
                for input_key in (
                    constants.INPUT_NAME_PATTERN.match(key)
                    for key in logic.dynamic_input_keys
                )
                if input_key
            )

        case Workflow() as workflow:
            return {
                parent_key.group("root")
                for parent_key in (
                    constants.PARENT_ROOT_PATTERN.match(key)
                    for key in workflow.dynamic_input_keys
                )
                if parent_key
            }

    # NOTE: This should never happen, but if it does set to empty set.
    return set()


def _step_label_error_diagnostic(
    label: str, step_semantic_block, semantic_anchor, message: str
) -> list[types.Diagnostic]:
    label_block = block_range_extract(
        search_key=f"label:{label}",
        search_nodes=step_semantic_block.children,
        anchor=semantic_anchor,
    )
    match label_block:
        case list(label_diagnostics):
            return label_diagnostics
        case None:
            diagnostic_range = compute_abs_range(step_semantic_block, semantic_anchor)
        case _:
            diagnostic_range = compute_abs_range(label_block, semantic_anchor)

    return [
        types.Diagnostic(
            message=message,
            severity=types.DiagnosticSeverity.Warning,
            range=diagnostic_range,
        )
    ]



from __future__ import annotations
import hashlib

from yaml.loader import SafeLoader
from yaml.nodes import Node


from .koreo_semantics import ALL, SEMANTIC_TYPE_STRUCTURE
from .semantics import (
    Position,
    SemanticAnchor,
    SemanticBlock,
    SemanticNode,
    SemanticStructure,
    TokenModifiers,
    TokenTypes,
    compute_abs_position,
    compute_abs_range,
)

from .extractor import extract_semantic_structure_info

STRUCTURE_KEY = "..structure.."


class IndexingLoader(SafeLoader):
    def __init__(self, *args, doc, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_node_abs_start = Position(line=0, character=0)
        self.last_node_abs_end = Position(line=0, character=0)
        self.doc = doc
        self.doc_count = 0

    def construct_document(self, node: Node):
        yaml_doc = super().construct_document(node)
        if not yaml_doc:
            self.doc_count = self.doc_count + 1
            return

        start_line = node.start_mark.line
        end_line = node.end_mark.line
        block_value = ("\n".join(self.doc.lines[start_line:end_line])).encode()
        block_hash = hashlib.md5(block_value, usedforsecurity=False).hexdigest()

        doc_kind = yaml_doc.get("kind")
        doc_semantics = SEMANTIC_TYPE_STRUCTURE.get(doc_kind)
        if not doc_semantics:
            doc_semantics = SEMANTIC_TYPE_STRUCTURE.get(ALL, SemanticStructure())

        doc_metadata = yaml_doc.get("metadata", {})
        doc_name = doc_metadata.get("name")

        if not doc_kind:
            anchor_key = f"Unknown:{self.doc_count}"
        elif doc_kind and doc_name:
            anchor_key = f"{doc_kind}:{doc_name}"
        else:
            anchor_key = f"{doc_kind}:{self.doc_count}"

        anchor_abs_start = Position(
            line=node.start_mark.line,
            character=node.start_mark.column,
        )

        anchor_rel_start = Position(
            line=node.start_mark.line - self.last_node_abs_start.line,
            character=node.start_mark.column,
        )

        structure, last_abs_start = extract_semantic_structure_info(
            anchor_abs_start=anchor_abs_start,
            last_token_abs_start=self.last_node_abs_start,
            yaml_node=node,
            doc=self.doc,
            semantic_type=doc_semantics,
        )
        yaml_doc[STRUCTURE_KEY] = SemanticAnchor(
            key=anchor_key,
            abs_position=anchor_abs_start,
            rel_position=anchor_rel_start,
            children=structure,
        )

        self.last_node_abs_start = last_abs_start

        self.doc_count = self.doc_count + 1

        return block_hash, yaml_doc



import random
import unittest


from koreo_tooling.indexing import semantics


class TestFlatten(unittest.TestCase):
    def test_no_children(self):
        expected_nodes = []

        def _node_factory():
            node_length = random.randint(5, 50)
            node_type = random.choice(semantics.TokenTypes)
            node_modifier = random.choices(
                [modifier for modifier in semantics.Modifier], k=random.randint(0, 3)
            )

            node = semantics.SemanticNode(
                position=semantics.Position(line=0, character=0),
                anchor_rel=semantics.Position(line=0, character=0),
                length=node_length,
                node_type=node_type,
                modifier=node_modifier,
                children=None,
            )

            expected_nodes.append(node)

            return node

        nodes = [_node_factory() for _ in range(random.randint(1, 10))]

        flattened = semantics.flatten(nodes)

        self.assertListEqual(expected_nodes, flattened)

    def test_children(self):

        def _node_factory(depth: int = 0):
            node_length = random.randint(5, 50)
            node_type = random.choice(semantics.TokenTypes)
            node_modifier = random.choices(
                [modifier for modifier in semantics.Modifier], k=random.randint(0, 3)
            )

            child_nodes = []
            children = None

            child_count = random.randint(0, depth)
            if child_count:
                children = []
                for _ in range(child_count):
                    child_node, child_expected_nodes = _node_factory(depth=depth - 1)
                    children.append(child_node)
                    child_nodes.extend(child_expected_nodes)

            node = semantics.SemanticNode(
                position=semantics.Position(line=0, character=0),
                anchor_rel=semantics.Position(line=0, character=0),
                length=node_length,
                node_type=node_type,
                modifier=node_modifier,
                children=children,
            )

            childless_node = semantics.SemanticNode(
                position=semantics.Position(line=0, character=0),
                anchor_rel=semantics.Position(line=0, character=0),
                length=node_length,
                node_type=node_type,
                modifier=node_modifier,
                children=None,
            )

            expected_nodes = [childless_node]
            expected_nodes.extend(child_nodes)

            return node, expected_nodes

        nodes = []
        expected_nodes = []
        for _ in range(1, random.randint(3, 10)):
            child_node, child_expected_nodes = _node_factory(5)
            nodes.append(child_node)
            expected_nodes.extend(child_expected_nodes)

        flattened = semantics.flatten(nodes)

        self.maxDiff = None
        self.assertListEqual(expected_nodes, flattened)


class TestFlattenNode(unittest.TestCase):
    def test_no_children(self):
        node_length = random.randint(5, 50)
        node_type = random.choice(semantics.TokenTypes)
        node_modifier = random.choices(
            [modifier for modifier in semantics.Modifier], k=random.randint(0, 3)
        )

        node = semantics.SemanticNode(
            position=semantics.Position(line=0, character=0),
            anchor_rel=semantics.Position(line=0, character=0),
            length=node_length,
            node_type=node_type,
            modifier=node_modifier,
            children=None,
        )

        flattened = semantics.flatten_node(node)

        self.assertListEqual(
            [
                semantics.SemanticNode(
                    position=semantics.Position(line=0, character=0),
                    anchor_rel=semantics.Position(line=0, character=0),
                    length=node_length,
                    node_type=node_type,
                    modifier=node_modifier,
                    children=None,
                )
            ],
            flattened,
        )

    def test_children(self):

        def _node_factory(depth: int = 0):
            node_length = random.randint(5, 50)
            node_type = random.choice(semantics.TokenTypes)
            node_modifier = random.choices(
                [modifier for modifier in semantics.Modifier], k=random.randint(0, 3)
            )

            child_nodes = []
            children = None

            child_count = random.randint(0, depth)
            if child_count:
                children = []
                for _ in range(child_count):
                    child_node, child_expected_nodes = _node_factory(depth=depth - 1)
                    children.append(child_node)
                    child_nodes.extend(child_expected_nodes)

            node = semantics.SemanticNode(
                position=semantics.Position(line=0, character=0),
                anchor_rel=semantics.Position(line=0, character=0),
                length=node_length,
                node_type=node_type,
                modifier=node_modifier,
                children=children,
            )

            childless_node = semantics.SemanticNode(
                position=semantics.Position(line=0, character=0),
                anchor_rel=semantics.Position(line=0, character=0),
                length=node_length,
                node_type=node_type,
                modifier=node_modifier,
                children=None,
            )

            expected_nodes = [childless_node]
            expected_nodes.extend(child_nodes)

            return node, expected_nodes

        child_node, expected_nodes = _node_factory(5)
        flattened = semantics.flatten_node(child_node)

        self.maxDiff = None
        self.assertListEqual(expected_nodes, flattened)
